/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul.h>

#include "component_info.h"
#include "private.h"

namespace component_manager {

ComponentInfo::LocalizedInfo::LocalizedInfo(std::string locale,
                                            std::string icon,
                                            std::string label)
  : locale_(std::move(locale)),
    icon_(std::move(icon)),
    label_(std::move(label)) {
}

ComponentInfo::LocalizedInfo::~LocalizedInfo() = default;

std::string ComponentInfo::LocalizedInfo::GetLocale() {
  return locale_;
}

std::string ComponentInfo::LocalizedInfo::GetIcon() {
  return icon_;
}

std::string ComponentInfo::LocalizedInfo::GetLabel() {
  return label_;
}

ComponentInfo::ComponentInfo(uid_t uid, std::string component_id)
  : handle_(nullptr, aul_compinfo_destroy), uid_(uid) {
  aul_compinfo_h handle_inst = nullptr;
  aul_compinfo_usr_create(component_id.c_str(), uid, &handle_inst);
  handle_.reset(handle_inst);
}

ComponentInfo::ComponentInfo(aul_compinfo_h info, uid_t uid)
  : handle_(info, aul_compinfo_destroy), uid_(uid) {
}

ComponentInfo::~ComponentInfo() = default;

std::string ComponentInfo::GetAppID() {
  const char* app_id = nullptr;
  aul_compinfo_get_app_id(handle_.get(), &app_id);
  if (app_id)
    return std::string(app_id);

  return std::string("");
}

std::string ComponentInfo::GetComponentID() {
  const char* comp_id = nullptr;
  aul_compinfo_get_comp_id(handle_.get(), &comp_id);
  if (comp_id)
    return std::string(comp_id);

  return std::string("");
}

std::string ComponentInfo::GetType() {
  const char* type = nullptr;
  aul_compinfo_get_type(handle_.get(), &type);
  if (type)
    return std::string(type);

  return std::string("");
}

std::string ComponentInfo::GetLaunchMode() {
  const char* launch_mode = nullptr;
  aul_compinfo_get_launch_mode(handle_.get(), &launch_mode);
  if (launch_mode)
    return std::string(launch_mode);

  return std::string("");
}

bool ComponentInfo::IsMainComponent() {
  bool is_main_comp = false;
  aul_compinfo_is_main_comp(handle_.get(), &is_main_comp);
  return is_main_comp;
}

bool ComponentInfo::IsIconDisplay() {
  bool is_icon_display = false;
  aul_compinfo_is_icon_display(handle_.get(), &is_icon_display);
  return is_icon_display;
}

bool ComponentInfo::IsManagedByTaskManager() {
  bool managed = false;
  aul_compinfo_is_taskmanage(handle_.get(), &managed);
  return managed;
}

std::vector<std::unique_ptr<ComponentInfo::LocalizedInfo>>
ComponentInfo::GetLocalizedInfos() {
  std::vector<std::unique_ptr<LocalizedInfo>> infos;
  aul_compinfo_usr_foreach_localized_info(GetComponentID().c_str(), uid_,
      [](aul_compinfo_localized_info_h handle, void* user_data) -> bool {
        const char* locale;
        aul_compinfo_localized_info_get_locale(handle, &locale);
        const char* icon;
        aul_compinfo_localized_info_get_icon(handle, &icon);
        if (icon == nullptr)
          icon = "";
        const char* label;
        aul_compinfo_localized_info_get_label(handle, &label);
        if (label == nullptr)
          label = "";

        auto* infos = static_cast<std::vector<std::unique_ptr<LocalizedInfo>>*>(
            user_data);
        infos->emplace_back(
            new (std::nothrow) LocalizedInfo(locale, icon, label));
        return true;
      }, &infos);
  return infos;
}

tizen_base::Bundle ComponentInfo::ToBundle() {
  tizen_base::Bundle b;
  b.Add(AUL_K_APPID, GetAppID());
  b.Add(AUL_K_COMPONENT_ID, GetComponentID());
  b.Add(AUL_K_COMPONENT_TYPE, GetType());
  b.Add(AUL_K_LAUNCH_MODE, GetLaunchMode());
  b.Add(AUL_K_MAIN_COMP, IsMainComponent() ? "true" : "false");
  b.Add(AUL_K_ICON_DISPLAY, IsIconDisplay() ? "true" : "false");
  b.Add(AUL_K_TASK_MANAGE, IsManagedByTaskManager() ? "true" : "false");

  std::vector<std::string> localized_info_arr;
  std::vector<std::string> localized_info_size_arr;
  for (auto& info : GetLocalizedInfos()) {
    tizen_base::Bundle localized_info_b;
    localized_info_b.Add(AUL_K_LOCALE, info->GetLocale());
    localized_info_b.Add(AUL_K_ICON, info->GetIcon());
    localized_info_b.Add(AUL_K_LABEL, info->GetLabel());

    auto raw = localized_info_b.ToRaw();
    localized_info_arr.emplace_back(reinterpret_cast<char*>(raw.first.get()));
    localized_info_size_arr.push_back(std::to_string(raw.second));
  }
  b.Add(AUL_K_LOCALIZED_INFO, localized_info_arr);
  b.Add(AUL_K_LOCALIZED_INFO_SIZE, localized_info_size_arr);
  return b;
}

}  // namespace component_manager
