/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SHARED_QUEUE_H_
#define SHARED_QUEUE_H_

#include <condition_variable>
#include <thread>
#include <mutex>
#include <memory>
#include <queue>

namespace component_manager {

template <class T>
class SharedQueue {
 public:
  SharedQueue() = default;
  virtual ~SharedQueue() = default;

  void Push(T item) {
    std::lock_guard<std::mutex> lock(mutex_);
    queue_.push(item);
    cond_var_.notify_one();
  }

  T WaitAndPop() {
    std::unique_lock<std::mutex> lock(mutex_);
    while (queue_.empty())
      cond_var_.wait(lock);

    auto item = std::move(queue_.front());
    queue_.pop();
    return item;
  }

  bool IsEmpty() const {
    std::lock_guard<std::mutex> lock(mutex_);
    return queue_.empty();
  }

  unsigned int Size() const {
    std::lock_guard<std::mutex> lock(mutex_);
    return queue_.size();
  }

 private:
  std::queue<T> queue_;
  mutable std::mutex mutex_;
  std::condition_variable cond_var_;
};

}  // namespace component_manager

#endif  // SHARED_QUEUE_H_
