/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <fcntl.h>

#include "worker.h"
#include "private.h"

namespace component_manager {

Worker::Worker(std::string name) : name_(std::move(name)) {
  thread_ = std::thread([&]{
        SetComm();
        Run();
      });
}

Worker::~Worker() {
  Quit();
  Join();
}

bool Worker::Join() {
  if (!thread_.joinable())
    return false;

  thread_.join();
  return true;
}

void Worker::Send(std::shared_ptr<Job> job) {
  queue_.Push(std::move(job));
}

void Worker::Run() {
  do {
    auto job = queue_.WaitAndPop();
    job->Do();
  } while (!done_);
}

unsigned int Worker::GetJobCount() const {
  return queue_.Size();
}

void Worker::Quit() {
  Send(std::shared_ptr<Job>(new (std::nothrow) Job("Quit", this)));
}

void Worker::SetComm() {
  pid_t tid = syscall(__NR_gettid);
  _I("[__WORKER__] Thread(%d)", tid);

  std::string path = "/proc/" + std::to_string(tid) + "/comm";
  int fd = open(path.c_str(), O_WRONLY);
  if (fd < 0) {
    _E("Failed to open %s. error(%d)", path.c_str(), errno);
    return;
  }

  ssize_t bytes_written = write(fd, name_.c_str(), name_.length() + 1);
  if (bytes_written < 0) {
    _E("Failed to write name(%s)", name_.c_str());
    close(fd);
    return;
  }

  close(fd);
}

void Worker::OnJob() {
  done_ = true;
}

}  // namespace component_manager
