/*
 * Copyright (c) 2017 - 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/types.h>
#include <unistd.h>

#include <glib.h>
#include <gio/gio.h>
#include <aul.h>
#include <aul_sock.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <rua_internal.h>
#include <rua_stat_internal.h>
#include <dlog.h>

#include <amd.h>

#include "rua_request.h"
#include "rua_info.h"
#include "rua_launch_context.h"
#include "rua_image_monitor.h"
#include "private.h"

#define AUL_SVC_K_LAUNCH_MODE "__APP_SVC_LAUNCH_MODE__"

typedef struct _rua_data_t {
	pid_t pid;
	uid_t uid;
	char *stat_tag;
	char *stat_caller;
	char *app_id;
	char *app_path;
	char *comp_id;
	char *instance_id;
	char *instance_name;
	char *icon;
	char *uri;
	bool is_group_mode;
	char *arg;
} rua_data_t;

typedef int (*rua_data_add_cb)(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status);
typedef void (*rua_data_remove_cb)(rua_data_t *rua);

typedef struct _rua_data_vft {
	rua_data_add_cb ctor;
	rua_data_remove_cb dtor;
} rua_data_vft;

struct rua_context_s {
	GList *user_list;
};

static struct rua_context_s *__get_context(void)
{
	static struct rua_context_s context;

	return &context;
}

static bool __is_group_mode(bundle *kb, uid_t uid)
{
	const char *str;
	const char *mode;
	const char *appid;
	amd_appinfo_h ai;

	if (kb == NULL)
		return false;

	appid = bundle_get_val(kb, AUL_K_APPID);
	if (appid == NULL)
		return false;

	ai = amd_appinfo_find(uid, appid);
	mode = amd_appinfo_get_value(ai, AMD_AIT_LAUNCH_MODE);
	if (mode != NULL && strcmp(mode, "caller") == 0) {
		str = bundle_get_val(kb, AUL_SVC_K_LAUNCH_MODE);
		if (str != NULL && strcmp(str, "group") == 0)
			return true;
	} else if (mode != NULL && strcmp(mode, "group") == 0) {
		return true;
	}

	return false;
}

static void __rua_data_remove_comp_id(rua_data_t *rua)
{
	if (!rua || !rua->comp_id)
		return;

	free(rua->comp_id);
	rua->comp_id = NULL;
}

static void __rua_data_remove_arg(rua_data_t *rua)
{
	if (!rua || !rua->arg)
		return;

	free(rua->arg);
	rua->arg = NULL;
}

static void __rua_data_remove_uri(rua_data_t *rua)
{
	if (!rua || !rua->uri)
		return;

	free(rua->uri);
	rua->uri = NULL;
}

static void __rua_data_remove_icon(rua_data_t *rua)
{
	if (!rua || !rua->icon)
		return;

	free(rua->icon);
	rua->icon = NULL;
}

static void __rua_data_remove_instance_name(rua_data_t *rua)
{
	if (!rua || !rua->instance_name)
		return;

	free(rua->instance_name);
	rua->instance_name = NULL;
}

static void __rua_data_remove_instance_id(rua_data_t *rua)
{
	if (!rua || !rua->instance_id)
		return;

	free(rua->instance_id);
	rua->instance_id = NULL;
}

static void __rua_data_remove_app_path(rua_data_t *rua)
{
	if (!rua || !rua->app_path)
		return;

	free(rua->app_path);
	rua->app_path = NULL;
}

static void __rua_data_remove_app_id(rua_data_t *rua)
{
	if (!rua || !rua->app_id)
		return;

	free(rua->app_id);
	rua->app_id = NULL;
}

static void __rua_data_remove_stat_caller(rua_data_t *rua)
{
	if (!rua || !rua->stat_caller)
		return;

	free(rua->stat_caller);
	rua->stat_caller = NULL;
}

static void __rua_data_remove_stat_tag(rua_data_t *rua)
{
	if (!rua || !rua->stat_tag)
		return;

	free(rua->stat_tag);
	rua->stat_tag = NULL;
}

static int __rua_data_add_comp_id(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	amd_app_type_e app_type;
	const char *comp_id;
	bundle *kb;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type != AMD_AT_COMPONENT_BASED_APP)
		return 0;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Failed to get bundle");
		return -1;
	}

	comp_id = bundle_get_val(kb, AUL_K_COMPONENT_ID);
	if (!comp_id) {
		_E("Failed to get component ID");
		return -1;
	}

	rua->comp_id = strdup(comp_id);
	if (!rua->comp_id) {
		_E("Failed to duplicate component ID");
		return -1;
	}

	return 0;
}

static int __rua_data_add_arg(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	if (amd_request_get_len(req) <= 0)
		return 0;

	rua->arg = calloc(amd_request_get_len(req) + 1, sizeof(char));
	if (!rua->arg) {
		_E("Out of memory");
		return -1;
	}

	memcpy(rua->arg, amd_request_get_raw(req), amd_request_get_len(req));

	return 0;
}

static int __rua_data_add_group_mode(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	bundle *kb = amd_request_get_bundle(req);
	uid_t uid = amd_request_get_target_uid(req);

	rua->is_group_mode = __is_group_mode(kb, uid);

	return 0;
}

static int __rua_data_add_uri(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *uri = _rua_launch_context_get_uri();

	rua->uri = strdup(uri ? uri : "");
	if (!rua->uri) {
		_E("Failed to duplicate uri");
		return -1;
	}

	return 0;
}

static int __rua_data_add_icon(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *icon = _rua_launch_context_get_icon();

	rua->icon = strdup(icon ? icon : "");
	if (!rua->icon) {
		_E("Failed to duplicate icon");
		return -1;
	}

	return 0;
}

static int __rua_data_add_instance_name(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *instance_name = _rua_launch_context_get_name();

	rua->instance_name = strdup(instance_name ? instance_name : "");
	if (!rua->instance_name) {
		_E("Failed to duplicate instance name");
		return -1;
	}

	return 0;
}

static int __rua_data_add_instance_id(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *inst_id = _rua_launch_context_get_id();
	amd_appinfo_h ai;
	const char *app_id;
	const char *multiple;
	uid_t uid;

	if (!inst_id) {
		uid = amd_app_status_get_uid(app_status);
		app_id = amd_app_status_get_appid(app_status);
		ai = amd_appinfo_find(uid, app_id);
		if (!ai) {
			_E("Failed to find appinfo");
			return -1;
		}

		multiple = amd_appinfo_get_value(ai, AMD_AIT_MULTI);
		if (multiple && !strcmp(multiple, "true"))
			inst_id = amd_app_status_get_instance_id(app_status);
	}

	if (inst_id) {
		rua->instance_id = strdup(inst_id);
		if (!rua->instance_id) {
			_E("Failed to duplicate instance ID");
			return -1;
		}
	}

	return 0;
}

static int __rua_data_add_app_path(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *app_path;

	app_path = amd_app_status_get_app_path(app_status);
	if (!app_path) {
		_E("Failed to get application path");
		return -1;
	}

	rua->app_path = strdup(app_path);
	if (!rua->app_path) {
		_E("Failed to duplicate application path");
		return -1;
	}

	return 0;
}

static int __rua_data_add_app_id(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *app_id;

	app_id = amd_app_status_get_appid(app_status);
	if (!app_id) {
		_E("Failed to get application ID");
		return -1;
	}

	rua->app_id = strdup(app_id);
	if (!rua->app_id) {
		_E("Failed to duplicate application ID");
		return -1;
	}

	return 0;
}

static int __rua_data_add_stat_caller(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *stat_caller;
	bundle *kb;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Failed to get bundle");
		return -1;
	}

	stat_caller = bundle_get_val(kb, AUL_SVC_K_RUA_STAT_CALLER);
	if (!stat_caller)
		return 0;

	rua->stat_caller = strdup(stat_caller);
	if (!rua->stat_caller) {
		_E("Failed to duplicate stat caller");
		return -1;
	}

	return 0;
}

static int __rua_data_add_stat_tag(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	const char *stat_tag;
	bundle *kb;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Failed to get bundle");
		return -1;
	}

	stat_tag = bundle_get_val(kb, AUL_SVC_K_RUA_STAT_TAG);
	if (!stat_tag)
		return 0;

	rua->stat_tag = strdup(stat_tag);
	if (!rua->stat_tag) {
		_E("Failed to duplicate stat tag");
		return -1;
	}

	return 0;
}

static int __rua_data_add_uid(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	rua->uid = amd_request_get_target_uid(req);

	return 0;
}

static int __rua_data_add_pid(rua_data_t *rua, amd_request_h req,
		amd_app_status_h app_status)
{
	rua->pid = amd_app_status_get_pid(app_status);

	return 0;
}

static rua_data_vft __rua_data_table[] = {
	{ __rua_data_add_pid, NULL },
	{ __rua_data_add_uid, NULL },
	{ __rua_data_add_stat_tag, __rua_data_remove_stat_tag },
	{ __rua_data_add_stat_caller, __rua_data_remove_stat_caller },
	{ __rua_data_add_app_id, __rua_data_remove_app_id },
	{ __rua_data_add_app_path, __rua_data_remove_app_path },
	{ __rua_data_add_instance_id, __rua_data_remove_instance_id },
	{ __rua_data_add_instance_name, __rua_data_remove_instance_name },
	{ __rua_data_add_icon, __rua_data_remove_icon },
	{ __rua_data_add_uri, __rua_data_remove_uri },
	{ __rua_data_add_group_mode, NULL },
	{ __rua_data_add_arg, __rua_data_remove_arg },
	{ __rua_data_add_comp_id, __rua_data_remove_comp_id },
};

static void __destroy_rua_data(void *data)
{
	rua_data_t *rua = data;
	int i;

	if (!rua)
		return;

	for (i = 0; i < ARRAY_SIZE(__rua_data_table); ++i) {
		if (__rua_data_table[i].dtor)
			__rua_data_table[i].dtor(rua);
	}

	free(rua);
}

static rua_data_t *__create_rua_data(amd_request_h req, pid_t pid)
{
	amd_app_status_h app_status;
	rua_data_t *rua;
	int ret;
	int i;

	if (!req) {
		_E("Invalid parameter");
		return NULL;
	}

	app_status = amd_app_status_find_by_pid(pid);
	if (!app_status) {
		_E("Failed to find app status. pid(%d)", pid);
		return NULL;
	}

	rua = calloc(1, sizeof(rua_data_t));
	if (!rua) {
		_E("Out of memory");
		return NULL;
	}

	for (i = 0; i < ARRAY_SIZE(__rua_data_table); ++i) {
		if (__rua_data_table[i].ctor) {
			ret = __rua_data_table[i].ctor(rua, req, app_status);
			if (ret < 0) {
				__destroy_rua_data(rua);
				return NULL;
			}
		}
	}

	return rua;
}

static gboolean __add_history_handler(gpointer user_data)
{
	rua_data_t *rua = (rua_data_t *)user_data;
	struct rua_rec rec = { 0, };
	const char *image;
	int ret;

	if (!rua)
		return G_SOURCE_REMOVE;

	if (!rua->is_group_mode) {
		rec.pkg_name = rua->app_id;
		rec.app_path = rua->app_path;
		rec.arg = rua->arg;
		rec.launch_time = time(NULL);
		rec.instance_id = rua->instance_id;
		rec.instance_name = rua->instance_name;
		rec.icon = rua->icon;
		rec.uri = rua->uri;

		image = _rua_info_get_image(rua->instance_id);
		if (image)
			rec.image = (char *)image;

		ret = rua_usr_db_add_history(&rec, rua->uid);
		if (ret < 0)
			_W("Failed to add rua history");

		SECURE_LOGD("RUA history added. app_id(%s) app_path(%s)",
				rec.pkg_name, rec.app_path);
	}

	if (rua->stat_caller && rua->stat_tag) {
		ret = rua_stat_usr_db_update(rua->stat_caller, rua->stat_tag,
				rua->uid);
		if (ret < 0)
			_W("Failed to update rua stat");

		SECURE_LOGD("RUA stat updated. caller(%s), tag(%s)",
				rua->stat_caller, rua->stat_tag);
	}

	__destroy_rua_data(rua);

	return G_SOURCE_REMOVE;
}

static bool __user_exist(uid_t uid)
{
	struct rua_context_s *context = __get_context();
	GList *iter;
	uid_t u;

	iter = context->user_list;
	while (iter) {
		u = GPOINTER_TO_UINT(iter->data);
		if (u == uid)
			return true;

		iter = g_list_next(iter);
	}

	return false;
}

static void __user_add(uid_t uid)
{
	struct rua_context_s *context = __get_context();

	context->user_list = g_list_append(context->user_list,
			GUINT_TO_POINTER(uid));
}

static void __user_remove(uid_t uid)
{
	struct rua_context_s *context = __get_context();

	context->user_list = g_list_remove(context->user_list,
			GUINT_TO_POINTER(uid));
}

static int __reply_foreach_cb(const char *key, void *user_data)
{
	if (key && !strcmp(key, "rua") && user_data) {
		g_timeout_add(1500, __add_history_handler, user_data);
		return 0;
	}

	return -1;
}

static int __on_request_user_init(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	uid_t uid = (uid_t)arg1;

	_I("user(%u)", uid);

	if (__user_exist(uid))
		return AMD_NOTI_CONTINUE;

	rua_usr_db_delete_history(NULL, uid);
	__user_add(uid);

	return AMD_NOTI_CONTINUE;
}

static int __on_login_monitor_user_logout(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	uid_t uid = (uid_t)arg1;

	_I("user(%u)", uid);

	__user_remove(uid);
	_rua_info_remove(NULL, uid);

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_app_start_pend(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	pid_t pid = (pid_t)arg1;
	bool bg_launch = (bool)arg2;
	amd_request_h req = (amd_request_h)arg3;
	amd_request_reply_h reply = (amd_request_reply_h)data;
	rua_data_t *rua = NULL;
	bundle *kb;
	int ret;

	if (!bg_launch) {
		kb = amd_request_get_bundle(req);
		ret = amd_noti_send(AMD_NOTI_MSG_RUA_SAVE_CRITICAL, 0, 0,
				req, kb);
		if (ret == 0) {
			rua = __create_rua_data(req, pid);
			if (rua && !rua->is_group_mode)
				_rua_info_add(kb, pid);
		}
	}

	_rua_launch_context_free();
	if (rua) {
		amd_request_reply_add_extra(reply, "rua",
				rua, __destroy_rua_data);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_app_start_end(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	pid_t pid = (pid_t)arg1;
	bool bg_launch = (bool)arg2;
	amd_request_h req = (amd_request_h)arg3;
	bundle *kb = data;
	rua_data_t *rua = NULL;
	int ret;

	if (pid > 0 && !bg_launch) {
		ret = amd_noti_send(AMD_NOTI_MSG_RUA_SAVE_CRITICAL, 0, 0,
				req, kb);
		if (ret == 0) {
			rua = __create_rua_data(req, pid);
			if (!rua) {
				_rua_launch_context_free();
				return AMD_NOTI_STOP;
			}

			g_timeout_add(1500, __add_history_handler, rua);
		}
	}

	_rua_launch_context_free();

	return AMD_NOTI_CONTINUE;
}

static void __rua_image_monitor_event_cb(const char *image, void *user_data)
{
	_rua_info_update_image(image);
}

static int __on_launch_app_startup_signal_end(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	int pid = arg1;

	amd_request_reply_foreach_extra(pid, __reply_foreach_cb);

	return AMD_NOTI_CONTINUE;
}

EXPORT int AMD_MOD_INIT(void)
{
	int ret;

	_D("RUA Initialize");

	ret = _rua_launch_context_init();
	if (ret < 0)
		return ret;

	ret = _rua_request_init();
	if (ret < 0)
		return ret;

	ret = _rua_info_init();
	if (ret < 0)
		return ret;

	_rua_image_monitor_set_event_cb(__rua_image_monitor_event_cb, NULL);

	ret = _rua_image_monitor_init();
	if (ret < 0)
		return ret;

	amd_noti_listen(AMD_NOTI_MSG_REQUEST_USER_INIT,
			__on_request_user_init);
	amd_noti_listen(AMD_NOTI_MSG_LOGIN_MONITOR_LOGOUT,
			__on_login_monitor_user_logout);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_APP_START_PEND,
			__on_launch_app_start_pend);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_APP_START_END,
			__on_launch_app_start_end);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_APP_STARTUP_SIGNAL_END,
			__on_launch_app_startup_signal_end);

	return 0;
}

EXPORT void AMD_MOD_FINI(void)
{
	_D("RUA Finalize");

	_rua_image_monitor_fini();
	_rua_info_fini();
	_rua_request_fini();
	_rua_launch_context_fini();
}
