/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <aul.h>
#include <aul_sock.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <rua_internal.h>
#include <rua_stat_internal.h>
#include <amd.h>

#include "rua_info.h"
#include "rua_image_monitor.h"
#include "private.h"

struct rua_info_s {
	char *key;
	pid_t pid;
	uid_t uid;
	char *app_id;
	char *app_path;
	char *instance_id;
	char *image_path;
	char *temp_path;
	char *comp_id;
	GList *wid_list;
};

struct rua_info_context_s {
	GHashTable *table;
};

typedef bool (*rua_info_foreach_remove_cb)(struct rua_info_s *info,
		void *user_data);

static struct rua_info_context_s *__get_context(void)
{
	static struct rua_info_context_s inst;

	return &inst;
}

static void __destroy_rua_info(gpointer data)
{
	struct rua_info_s *info = (struct rua_info_s *)data;

	if (info == NULL)
		return;

	if (info->wid_list)
		g_list_free(info->wid_list);

	if (info->temp_path) {
		_rua_image_monitor_clear(info->temp_path);
		free(info->temp_path);
	}

	if (info->image_path) {
		_rua_image_monitor_clear(info->image_path);
		free(info->image_path);
	}

	if (info->comp_id)
		free(info->comp_id);

	if (info->instance_id)
		free(info->instance_id);

	if (info->app_path)
		free(info->app_path);

	if (info->app_id)
		free(info->app_id);

	if (info->key)
		free(info->key);

	free(info);
}

static struct rua_info_s *__create_rua_info(const char *key,
		const char *app_id, const char *app_path,
		const char *instance_id, const char *comp_id,
		pid_t pid, uid_t uid)
{
	struct rua_info_s *info;

	info = calloc(1, sizeof(struct rua_info_s));
	if (info == NULL) {
		_E("Out of memory");
		return NULL;
	}

	info->key = strdup(key);
	if (!info->key) {
		_E("Failed to duplicate key");
		__destroy_rua_info(info);
		return NULL;
	}

	info->app_id = strdup(app_id);
	if (!info->app_id) {
		_E("Failed to duplicate application ID");
		__destroy_rua_info(info);
		return NULL;
	}

	info->app_path = strdup(app_path);
	if (!info->app_path) {
		_E("Failed to duplicate application path");
		__destroy_rua_info(info);
		return NULL;
	}

	if (instance_id) {
		info->instance_id = strdup(instance_id);
		if (!info->instance_id) {
			_E("Failed to duplicate instance ID");
			__destroy_rua_info(info);
			return NULL;
		}
	}

	if (comp_id) {
		info->comp_id = strdup(comp_id);
		if (!info->comp_id) {
			_E("Failed to duplicate component ID");
			__destroy_rua_info(info);
			return NULL;
		}
	}

	info->pid = pid;
	info->uid = uid;

	return info;
}

static gint __compare_wid(gconstpointer a, gconstpointer b)
{
	int wid1 = GPOINTER_TO_INT(a);
	int wid2 = GPOINTER_TO_INT(b);

	if (wid1 == wid2)
		return 0;

	return -1;
}

static struct rua_info_s *__rua_info_find_by_wid(int wid)
{
	struct rua_info_context_s *context = __get_context();
	struct rua_info_s *info;
	GHashTableIter iter;
	gpointer key;
	gpointer value;
	GList *found;

	g_hash_table_iter_init(&iter, context->table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		info = (struct rua_info_s *)value;
		if (!info || !info->wid_list)
			continue;

		found = g_list_find_custom(info->wid_list, GINT_TO_POINTER(wid),
				__compare_wid);
		if (found)
			return info;
	}

	return NULL;
}

static struct rua_info_s *__rua_info_find_by_instance_id(const char *id)
{
	struct rua_info_context_s *context = __get_context();
	struct rua_info_s *info;

	info = (struct rua_info_s *)g_hash_table_lookup(context->table, id);
	if (info)
		return info;

	return NULL;
}

static bool __has_taskmanage(amd_app_status_h app_status, bundle *kb, uid_t uid)
{
	amd_app_type_e app_type;
	amd_compinfo_h ci;
	amd_appinfo_h ai;
	const char *taskmanage;
	const char *comp_id;
	const char *app_id;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_COMPONENT_BASED_APP) {
		comp_id = bundle_get_val(kb, AUL_K_COMPONENT_ID);
		if (!comp_id) {
			_E("Failed to get component ID");
			return false;
		}

		ci = amd_compinfo_find(uid, comp_id);
		if (!ci) {
			_E("Failed to find compinfo");
			return false;
		}

		taskmanage = amd_compinfo_get_value(ci,
				AMD_COMPINFO_TYPE_TASKMANAGE);
	} else {
		app_id = amd_app_status_get_appid(app_status);
		if (!app_id) {
			_E("Failed to get application ID");
			return false;
		}

		ai = amd_appinfo_find(uid, app_id);
		if (!ai) {
			_E("Failed to find appinfo");
			return false;
		}

		taskmanage = amd_appinfo_get_value(ai, AMD_AIT_TASKMANAGE);
	}

	if (taskmanage && !strcmp(taskmanage, "true"))
		return true;

	return false;
}

static bool __can_have_window(amd_app_status_h app_status,
		const char *instance_id)
{
	amd_comp_status_h comp_status;
	amd_comp_type_e comp_type;
	amd_app_type_e app_type;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_UI_APP)
		return true;

	if (app_type == AMD_AT_COMPONENT_BASED_APP) {
		comp_status = amd_comp_status_find_by_instance_id(instance_id);
		if (!comp_status) {
			_W("Failed to find comp status");
			return false;
		}

		comp_type = amd_comp_status_get_comp_type(comp_status);
		if (comp_type == AMD_CT_FRAME_COMP)
			return true;
	}

	return false;
}

static gboolean __foreach_remove_by_rua_info(gpointer key, gpointer value,
		gpointer user_data)
{
	struct rua_info_s *info = (struct rua_info_s *)value;
	struct rua_info_s *new_info = (struct rua_info_s *)user_data;

	if (info->uid == new_info->uid &&
			!strcmp(info->app_id, new_info->app_id)) {
		if (info->instance_id && new_info->instance_id &&
			!strcmp(info->instance_id, new_info->instance_id))
			return TRUE;
		else if (!info->instance_id && !new_info->instance_id)
			return TRUE;
	}

	return FALSE;
}

static gboolean __foreach_remove_by_uid(gpointer key, gpointer value,
		gpointer data)
{
	struct rua_info_s *info = (struct rua_info_s *)value;
	uid_t uid = GPOINTER_TO_UINT(data);

	if (info->uid == uid)
		return TRUE;

	return FALSE;
}

static gboolean __foreach_remove_by_app_id(gpointer key, gpointer value,
		gpointer data)
{
	struct rua_info_s *info = (struct rua_info_s *)value;
	const char *app_id = (const char *)data;

	if (!strcmp(info->app_id, app_id))
		return TRUE;

	return FALSE;
}

static gboolean __foreach_remove_by_app_path(gpointer key, gpointer value,
		gpointer data)
{
	struct rua_info_s *info = (struct rua_info_s *)value;
	const char *app_path = (const char *)data;

	if (!strcmp(info->app_path, app_path))
		return TRUE;

	return FALSE;
}

static gboolean __foreach_remove_by_instance_id(gpointer key, gpointer value,
		gpointer data)
{
	struct rua_info_s *info = (struct rua_info_s *)value;
	const char *instance_id = (const char *)data;

	if (info->instance_id && !strcmp(info->instance_id, instance_id))
		return TRUE;

	return FALSE;
}

int _rua_info_add(bundle *kb, pid_t pid)
{
	struct rua_info_context_s *context = __get_context();
	struct rua_info_s *info;
	amd_app_status_h app_status;
	amd_app_type_e app_type;
	const char *instance_id;
	const char *app_path;
	const char *comp_id;
	const char *app_id;
	const char *key;
	uid_t uid;

	app_status = amd_app_status_find_by_pid(pid);
	if (!app_status) {
		_E("Failed to find app status. pid(%d)", pid);
		return -1;
	}

	instance_id = bundle_get_val(kb, AUL_K_INSTANCE_ID);
	if (instance_id)
		key = instance_id;
	else
		key = amd_app_status_get_instance_id(app_status);

	if (!key) {
		_E("key is nullptr");
		return -1;
	}

	if (!__can_have_window(app_status, instance_id)) {
		_D("The instance cannot have a window");
		return 0;
	}

	uid = amd_app_status_get_uid(app_status);
	if (!__has_taskmanage(app_status, kb, uid)) {
		_W("The instance isn't managed by taskmanager");
		return 0;
	}

	app_path = amd_app_status_get_app_path(app_status);
	app_id = amd_app_status_get_appid(app_status);
	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_COMPONENT_BASED_APP)
		comp_id = bundle_get_val(kb, AUL_K_COMPONENT_ID);
	else
		comp_id = NULL;

	info = __create_rua_info(key, app_id, app_path, instance_id, comp_id,
			pid, uid);
	if (!info)
		return -1;

	g_hash_table_foreach_remove(context->table,
			__foreach_remove_by_rua_info, info);

	if (g_hash_table_contains(context->table, key))
		g_hash_table_replace(context->table, info->key, info);
	else
		g_hash_table_insert(context->table, info->key, info);

	return 0;
}

int _rua_info_remove(bundle *kb, uid_t uid)
{
	struct rua_info_context_s *context = __get_context();
	const char *instance_id = NULL;
	const char *app_path = NULL;
	const char *app_id = NULL;

	if (kb) {
		app_id = bundle_get_val(kb, AUL_K_RUA_PKGNAME);
		app_path = bundle_get_val(kb, AUL_K_RUA_APPPATH);
		instance_id = bundle_get_val(kb, AUL_K_RUA_INSTANCE_ID);
	}

	if (app_id) {
		g_hash_table_foreach_remove(context->table,
				__foreach_remove_by_app_id,
				(void *)app_id);
	} else if (app_path) {
		g_hash_table_foreach_remove(context->table,
				__foreach_remove_by_app_path,
				(void *)app_path);
	} else if (instance_id) {
		g_hash_table_foreach_remove(context->table,
				__foreach_remove_by_instance_id,
				(void *)instance_id);
	} else {
		g_hash_table_foreach_remove(context->table,
				__foreach_remove_by_uid,
				GUINT_TO_POINTER(uid));
		_rua_image_monitor_clear_all();
	}

	return 0;
}

int _rua_info_update_image(const char *image)
{
	struct rua_info_s *info;
	char *file;
	int pid = -1;
	int wid = 0;
	int num = -1;
	int ret;

	file = basename(image);
	if (!file) {
		_E("Failed to get basename. image(%s)", image);
		return -1;
	}

	ret = sscanf(file, "win_%d_%d-%d.png", &pid, &wid, &num);
	if (ret != 3) {
		_E("Failed to parse image file");
		return -1;
	}

	info = __rua_info_find_by_wid(wid);
	if (!info) {
		_E("Failed to find rua info. %s will be deleted", image);
		_rua_image_monitor_clear(image);
		return -1;
	}

	if (info->image_path) {
		if (!strcmp(info->image_path, image))
			return 0;

		if (info->temp_path && strcmp(info->temp_path, image) != 0) {
			_rua_image_monitor_clear(info->temp_path);
			free(info->temp_path);
		}

		info->temp_path = info->image_path;
	}

	ret = rua_usr_db_update_image(info->app_id, info->comp_id,
			info->instance_id, image, info->uid);
	if (ret < 0) {
		SECURE_LOGW("Failed to update image path(%s). instance(%s)",
				image, info->instance_id);
	}

	info->image_path = strdup(image);
	if (!info->image_path) {
		_E("Failed to duplicate image path(%s)", image);
		return -1;
	}
	SECURE_LOGI("[__RUA_INFO__] Updated. instance(%s), update(%s)",
			info->instance_id, info->image_path);

	return 0;
}

const char *_rua_info_get_image(const char *instance_id)
{
	struct rua_info_s *info;

	if (!instance_id)
		return NULL;

	info = __rua_info_find_by_instance_id(instance_id);
	if (!info)
		return NULL;

	return info->image_path;
}

static int __on_app_group_window_set(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int pid = arg1;
	int wid = arg2;
	const char *instance_id = (const char *)instance_id;
	amd_app_status_h app_status;
	amd_comp_status_h comp_status;
	const char *leader_id;
	struct rua_info_s *info;

	if (instance_id) {
		comp_status = amd_comp_status_find_by_instance_id(instance_id);
		if (!comp_status) {
			_W("Failed to find comp status");
			return AMD_NOTI_STOP;
		}

		leader_id = amd_comp_status_get_leader_id(comp_status);
	} else {
		app_status = amd_app_status_find_by_pid(pid);
		if (!app_status) {
			_W("Failed to find app status. pid(%d)", pid);
			return AMD_NOTI_STOP;
		}

		instance_id = amd_app_status_get_instance_id(app_status);
		leader_id = amd_app_status_get_leader_id(app_status);
	}

	if (leader_id)
		instance_id = leader_id;

	info = __rua_info_find_by_instance_id(instance_id);
	if (!info) {
		SECURE_LOGW("Failed to find rua info. instance(%s)",
				instance_id);
		return AMD_NOTI_CONTINUE;
	}

	info->wid_list = g_list_append(info->wid_list, GINT_TO_POINTER(wid));
	SECURE_LOGI("[__RUA_INFO__] instance(%s), wid(%d)", instance_id, wid);

	return AMD_NOTI_CONTINUE;
}

static int __on_app_group_destroy_app_group_context(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	int wid = arg2;
	struct rua_info_s *info;

	info = __rua_info_find_by_wid(wid);
	if (!info)
		return AMD_NOTI_CONTINUE;

	info->wid_list = g_list_remove(info->wid_list, GINT_TO_POINTER(wid));

	return AMD_NOTI_CONTINUE;
}

int _rua_info_init(void)
{
	struct rua_info_context_s *context = __get_context();

	_D("RUA Info Initialize");

	context->table = g_hash_table_new_full(g_str_hash, g_str_equal,
			NULL, __destroy_rua_info);
	if (!context->table) {
		_E("Out of memory");
		return -1;
	}

	amd_noti_listen(AMD_NOTI_MSG_APP_GROUP_WINDOW_SET,
			__on_app_group_window_set);
	amd_noti_listen(AMD_NOTI_MSG_APP_GROUP_DESTROY_APP_GROUP_CONTEXT,
			__on_app_group_destroy_app_group_context);

	return 0;
}

void _rua_info_fini(void)
{
	struct rua_info_context_s *context = __get_context();

	_D("RUA Info Finalize");

	if (context->table) {
		g_hash_table_destroy(context->table);
		context->table = NULL;
	}
}
