/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <gio/gio.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <rua_internal.h>
#include <rua_stat_internal.h>
#include <aul.h>
#include <amd.h>

#include "rua_launch_context.h"
#include "private.h"

#define MULTI_INSTANCE_SHORTCUT "multi-instance-shortcut"
#define QUERY_KEY_ID            "id="
#define QUERY_KEY_ICON          "icon="
#define QUERY_KEY_NAME          "name="
#define AUL_SVC_K_URI           "__APP_SVC_URI__"

struct rua_launch_context_s {
	char *id;
	char *name;
	char *icon;
	char *uri;
};

static struct rua_launch_context_s *__get_launch_context(void)
{
	static struct rua_launch_context_s context;

	return &context;
}

const char *_rua_launch_context_get_id(void)
{
	struct rua_launch_context_s *info = __get_launch_context();

	return info->id;
}

const char *_rua_launch_context_get_name(void)
{
	struct rua_launch_context_s *info = __get_launch_context();

	return info->name;
}

const char *_rua_launch_context_get_icon(void)
{
	struct rua_launch_context_s *info = __get_launch_context();

	return info->icon;
}

const char *_rua_launch_context_get_uri(void)
{
	struct rua_launch_context_s *info = __get_launch_context();

	return info->uri;
}

void _rua_launch_context_free(void)
{
	struct rua_launch_context_s *info = __get_launch_context();

	if (info->uri) {
		free(info->uri);
		info->uri = NULL;
	}

	if (info->id) {
		free(info->id);
		info->id = NULL;
	}

	if (info->icon) {
		free(info->icon);
		info->icon = NULL;
	}

	if (info->name) {
		free(info->name);
		info->name = NULL;
	}
}

static char *__get_value_from_query(const char *src, const char *key)
{
	int src_len = strlen(src);
	int key_len = strlen(key);

	if (src_len > key_len) {
		if (strncmp(src, key, key_len) == 0)
			return g_uri_unescape_string(src + key_len, NULL);
	}

	return NULL;
}

static int __set_launch_context(bundle *kb)
{
	struct rua_launch_context_s *info = __get_launch_context();
	const char *uri;
	gchar *scheme;
	gchar *query;
	char *token;
	char *saveptr = NULL;
	char *dup_uri;

	uri = bundle_get_val(kb, AUL_SVC_K_URI);
	if (uri == NULL)
		return -1;

	scheme = g_uri_parse_scheme(uri);
	if (scheme == NULL)
		return -1;

	if (strcmp(scheme, MULTI_INSTANCE_SHORTCUT) != 0) {
		g_free(scheme);
		return -1;
	}
	g_free(scheme);

	dup_uri = strdup(uri);
	if (dup_uri == NULL) {
		_E("Out of memory");
		return -1;
	}

	info->uri = strdup(uri);
	if (info->uri == NULL) {
		_E("Out of memory");
		free(dup_uri);
		return -1;
	}

	query = index(dup_uri, '?');
	if (query == NULL) {
		_rua_launch_context_free();
		free(dup_uri);
		return -1;
	}

	token = strtok_r(query + 1, "&", &saveptr);
	while (token != NULL) {
		if (info->id == NULL) {
			info->id = __get_value_from_query(token,
					QUERY_KEY_ID);
		}

		if (info->icon == NULL) {
			info->icon = __get_value_from_query(token,
					QUERY_KEY_ICON);
		}

		if (info->name == NULL) {
			info->name = __get_value_from_query(token,
					QUERY_KEY_NAME);
		}

		token = strtok_r(NULL, "&", &saveptr);
	}
	free(dup_uri);

	if (info->id == NULL || info->icon == NULL || info->name == NULL) {
		_W("Failed to get instance info");
		_rua_launch_context_free();
		return -1;
	}

	return 0;
}

int _rua_launch_context_set(bundle *kb)
{
	struct rua_launch_context_s *info = __get_launch_context();
	const char *instance_id;
	int ret;

	ret = __set_launch_context(kb);
	if (ret == 0) {
		bundle_del(kb, AUL_K_INSTANCE_ID);
		bundle_add(kb, AUL_K_INSTANCE_ID, info->id);
		bundle_del(kb, AUL_K_NEW_INSTANCE);
		bundle_add(kb, AUL_K_NEW_INSTANCE, "true");
		_D("Multiple instance launch - id(%s), name(%s), icon(%s)",
				info->id, info->name, info->icon);
		return 0;
	}

	instance_id = bundle_get_val(kb, AUL_K_INSTANCE_ID);
	if (!instance_id)
		return 0;

	info->id = strdup(instance_id);
	if (!info->id) {
		_W("Out of memory");
		return -1;
	}

	_D("Multiple instance launch - id(%s)", info->id);

	return 0;
}

static int __on_launch_app_start_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	_rua_launch_context_free();
	_rua_launch_context_set(data);

	return AMD_NOTI_CONTINUE;
}

int _rua_launch_context_init(void)
{
	_D("RUA Launch Context Initialize");

	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_APP_START_START,
			__on_launch_app_start_start);

	return 0;
}

void _rua_launch_context_fini(void)
{
	_D("RUA Launch Context Finalize");

	_rua_launch_context_free();
}
