/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <amd.h>
#include <aul.h>
#include <aul_cmd.h>
#include <bundle_internal.h>
#include <tizen-extension-client-protocol.h>
#include <wayland-client.h>
#include <wayland-tbm-client.h>
#include <tzplatform_config.h>
#include <tizen-launch-client-protocol.h>

#include "screen-resolution-private.h"

#define METADATA_KEY_BASE_SCREEN_RESOLUTION \
	"http://tizen.org/metadata/app_ui_type/base_screen_resolution"
#define DEFAULT_UID tzplatform_getuid(TZ_SYS_DEFAULT_USER)
#define TAG_RESOLUTION  "[Screen Resolution]"
#define PATH_CONF       "/usr/share/amd/conf/amd_screen_resolution.conf"

struct wayland_context_s {
	struct wl_display *display;
	struct tizen_launch_appinfo *launch_appinfo;
	uint32_t launch_appinfo_id;
};

static struct wayland_context_s __context;
static GList *__values;

static int __parse_values(const char *path)
{
        FILE *fp;
        char buf[LINE_MAX];
        char *token = NULL;

        fp = fopen(path, "rt");
        if (!fp) {
                LOGE("Failed to open %s", path);
                return -1;
        }

        while (fgets(buf, sizeof(buf), fp) != NULL) {
                sscanf(buf, "%ms", &token);
                if (!token)
                        continue;

                if (token[0] == '\0' || !strcasecmp(token, TAG_RESOLUTION)) {
                        free(token);
                        token = NULL;
                        continue;
                }

                __values = g_list_append(__values, token);
                token = NULL;
        }

        fclose(fp);

        if (__values == NULL) {
                LOGE("value list is empty");
                return -1;
        }

        return 0;
}

static int __launch_appinfo_register_app_pid(int pid)
{
	if (!__context.display) {
		__context.display = amd_wayland_get_display();
		if (!__context.display) {
			LOGW("Failed to get wayland display");
			return -1;
		}
	}

	if (!__context.launch_appinfo) {
		LOGW("Failed to bind tizen launch appinfo");
		return -1;
	}

	tizen_launch_appinfo_register_pid(__context.launch_appinfo,
			(uint32_t)pid);
	wl_display_flush(__context.display);
	LOGI("Register app pid: %d", pid);

	return 0;
}

static int __launch_appinfo_deregister_app_pid(int pid)
{
	if (!__context.display) {
		__context.display = amd_wayland_get_display();
		if (!__context.display) {
			LOGW("Failed to get wayland display");
			return -1;
		}
	}

	if (!__context.launch_appinfo) {
		LOGW("Failed to bind tizen launch appinfo");
		return -1;
	}

	tizen_launch_appinfo_deregister_pid(__context.launch_appinfo,
			(uint32_t)pid);
	wl_display_flush(__context.display);
	LOGI("Deregister app pid: %d", pid);

	return 0;
}

static int __metadata_foreach_cb(const char *value, void *user_data)
{
	bool *need_resolution = (bool *)user_data;
	GList *found;

	if (!value)
		return 0;

	found = g_list_find_custom(__values, value, (GCompareFunc)strcmp);
	if (found) {
		*need_resolution = true;
		return -1;
	}

	return 0;
}

static bool __need_screen_resolution(const char *appid)
{
	amd_app_property_h app_property;
	bool need_resolution = false;

	app_property = amd_app_property_find(DEFAULT_UID);
	if (!app_property)
		return false;

	amd_app_property_metadata_foreach(app_property, appid,
			METADATA_KEY_BASE_SCREEN_RESOLUTION,
			__metadata_foreach_cb,
			(void *)&need_resolution);

	return need_resolution;
}

static int __launch_appinfo_launch_app(const char *appid, int pid)
{
	if (!appid || pid < 0) {
		LOGE("Invalid parameter");
		return -1;
	}

	if (!__need_screen_resolution(appid))
		return 0;

	if (!__context.display) {
		__context.display = amd_wayland_get_display();
		if (!__context.display) {
			LOGW("Failed to get wayland display");
			return -1;
		}
	}

	if (!__context.launch_appinfo) {
		LOGW("Failed to bind tizen launch appinfo");
		return -1;
	}

	tizen_launch_appinfo_set_appid(__context.launch_appinfo,
			pid, appid);
	wl_display_flush(__context.display);
	LOGI("Launch app. %s:%d", appid, pid);

	return 0;
}

static int __on_wayland_listener_tizen_launch_appinfo(const char *msg,
		int arg1, int arg2, void *arg3, bundle *data)
{
	uint32_t id = (uint32_t)arg1;
	struct wl_registry *registry = (struct wl_registry *)arg3;

	if (!__context.launch_appinfo) {
		__context.launch_appinfo_id = id;
		__context.launch_appinfo = wl_registry_bind(registry, id,
				&tizen_launch_appinfo_interface, 1);
		LOGI("tizen_launch_appinfo(%p), id(%u)",
				__context.launch_appinfo, id);
	}

	if (!__context.display)
		__context.display = amd_wayland_get_display();

	return AMD_NOTI_CONTINUE;
}

static int __on_wayland_listener_remove(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	uint32_t id = (uint32_t)arg1;

	if (id == __context.launch_appinfo_id && __context.launch_appinfo) {
		tizen_launch_appinfo_destroy(__context.launch_appinfo);
		__context.launch_appinfo = NULL;
		__context.launch_appinfo_id = 0;
		LOGW("tizen_launch_appinfo is destroyed");
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_do_starting_app_end(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_launch_context_h h = (amd_launch_context_h)arg3;
	amd_appinfo_h ai = amd_launch_context_get_appinfo(h);
	int pid = amd_launch_context_get_pid(h);
	const char *comp_type;
	const char *appid;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		appid = amd_appinfo_get_value(ai, AMD_AIT_NAME);
		if (__need_screen_resolution(appid))
			__launch_appinfo_register_app_pid(pid);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_complete_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int pid = arg1;
	bool new_process = (bool)arg2;
	amd_appinfo_h ai = (amd_appinfo_h)arg3;
	const char *comp_type;
	const char *appid;

	if (!new_process)
		return AMD_NOTI_CONTINUE;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		appid = amd_appinfo_get_value(ai, AMD_AIT_NAME);
		__launch_appinfo_launch_app(appid, pid);
	} else {
		__launch_appinfo_deregister_app_pid(pid);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_main_app_dead_loader(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int pid = arg1;

	__launch_appinfo_deregister_app_pid(pid);

	return AMD_NOTI_CONTINUE;
}

static int __dispatch_launchpad_child_process(amd_request_h req)
{
	pid_t pid = amd_request_get_pid(req);

	__launch_appinfo_register_app_pid(pid);
	LOGI("Loader pid: %d", pid);

	return 0;
}

static amd_request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = LAUNCHPAD_CHILD_PROCESS,
		.callback = __dispatch_launchpad_child_process
	},
};

static amd_cynara_checker __cynara_checkers[] = {
	{
		.cmd = LAUNCHPAD_CHILD_PROCESS,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM,
		.priority = 10
	},
};

EXPORT int AMD_MOD_INIT(void)
{
	int r;

	LOGD("screen resolution init");

	r = amd_request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (r < 0) {
		LOGE("Failed to register dispatch table");
		return -1;
	}

	r = amd_cynara_register_checkers(__cynara_checkers,
			ARRAY_SIZE(__cynara_checkers));
	if (r < 0) {
		LOGE("Failed to register cynara checkers");
		return -1;
	}

	r = amd_app_property_metadata_add_filter(
			METADATA_KEY_BASE_SCREEN_RESOLUTION,
			NULL);
	if (r < 0) {
		LOGE("Failed to add metadata filter");
		return -1;
	}

	r = __parse_values(PATH_CONF);
	if (r < 0)
		LOGW("Failed to parse metadata value");

	amd_noti_listen(AMD_NOTI_MSG_WAYLAND_LISTENER_TIZEN_LAUNCH_APPINFO,
			__on_wayland_listener_tizen_launch_appinfo);
	amd_noti_listen(AMD_NOTI_MSG_WAYLAND_LISTENER_REMOVE,
			__on_wayland_listener_remove);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_DO_STARTING_APP_END,
			__on_launch_do_starting_app_end);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_COMPLETE_START,
			__on_launch_complete_start);
	amd_noti_listen(AMD_NOTI_MSG_MAIN_APP_DEAD_LOADER,
			__on_main_app_dead_loader);

	return 0;
}

EXPORT void AMD_MOD_FINI(void)
{
	LOGD("screen resolution fini");

	g_list_free_full(__values, free);
}

