/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <glib.h>
#include <gio/gio.h>
#include <aul.h>
#include <aul_cmd.h>
#include <aul_watchdog.h>
#include <aul_sock.h>
#include <bundle_internal.h>
#include <amd.h>

#include "amd_watchdog_private.h"
#include "amd_watchdog_config.h"
#include "amd_watchdog_logger.h"

#define ARRAY_SIZE(x) ((sizeof(x)) / sizeof(x[0]))

typedef struct watchdog_s {
	GDBusConnection *conn;
	guint listener;
	guint retry_handler;
	unsigned int interval;
	int max_retry_count;
	GList *proc_contexts;
} watchdog;

typedef struct proc_context_s {
	pid_t pid;
	uid_t uid;
	bool frozen;
	bool watchdog_enable;
	guint timer;
	int retry_count;
	int kick_count;
	struct timespec resp_time;
} proc_context;

static watchdog __watchdog;

static proc_context *__create_proc_context(pid_t pid, uid_t uid)
{
	proc_context *ctx;

	ctx = calloc(1, sizeof(proc_context));
	if (!ctx) {
		_E("Out of memory");
		return NULL;
	}

	ctx->pid = pid;
	ctx->uid = uid;

	return ctx;
}

static void __destroy_proc_context(gpointer data)
{
	proc_context *ctx = (proc_context *)data;

	if (ctx->timer > 0)
		g_source_remove(ctx->timer);

	free(ctx);
}

static proc_context *__find_proc_context(pid_t pid)
{
	proc_context *ctx;
	GList *iter;

	iter = __watchdog.proc_contexts;
	while (iter) {
		ctx = (proc_context *)iter->data;
		if (ctx && ctx->pid == pid)
			return ctx;

		iter = g_list_next(iter);
	}

	return NULL;
}

static const char *__get_appid(pid_t pid)
{
	amd_app_status_h app_status;
	const char *appid = NULL;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (app_status)
		appid = amd_app_status_get_appid(app_status);

	return appid ? appid : "Unknown";
}

static bundle *__create_bundle(unsigned int interval)
{
	char buf[32];
	bundle *b;
	int ret;

	b = bundle_create();
	if (!b) {
		_E("Failed to create bundle");
		return NULL;
	}

	if (interval != 0) {
		snprintf(buf, sizeof(buf), "%u", interval);
		ret = bundle_add(b, AUL_K_INTERVAL, buf);
		if (ret != BUNDLE_ERROR_NONE) {
			_E("Failed to add interval(%s)", buf);
			bundle_free(b);
			return NULL;
		}
	}

	return b;
}

static int __send_request(proc_context *ctx, int cmd, unsigned int interval)
{
	bundle *b;
	int ret;

	b = __create_bundle(interval);
	if (!b)
		return -ENOMEM;

	ret = aul_sock_send_bundle(ctx->pid, ctx->uid, cmd, b,
			AUL_SOCK_NOREPLY);
	bundle_free(b);
	if (ret < 0) {
		_E("Failed to send request(%d), pid(%d)", cmd, ctx->pid);
		return ret;
	}

	return 0;
}

static gboolean __timeout_handler(gpointer data)
{
	proc_context *ctx = (proc_context *)data;

	_E("Timer expired. pid(%d), resp_time(%ld.%ld)",
			ctx->pid,
			ctx->resp_time.tv_sec,
			ctx->resp_time.tv_nsec);

	if (ctx->retry_count < __watchdog.max_retry_count) {
		ctx->retry_count++;
		return G_SOURCE_CONTINUE;
	}

	_E("pid(%d) will be terminated", ctx->pid);
	amd_signal_send_watchdog(ctx->pid, SIGKILL);
	_watchdog_logger_print("SIGKILL", "pid(%d), appid(%s)",
			ctx->pid, __get_appid(ctx->pid));

	return G_SOURCE_REMOVE;
}

static void __watchdog_set_timer(proc_context *ctx)
{
	if (!ctx->timer) {
		ctx->timer = g_timeout_add(__watchdog.interval + 1000,
				__timeout_handler, ctx);
	}
}

static void __watchdog_unset_timer(proc_context *ctx)
{
	if (ctx->timer) {
		g_source_remove(ctx->timer);
		ctx->timer = 0;
	}
}

static int __dispatch_watchdog_enable(amd_request_h req)
{
	int pid = amd_request_get_pid(req);
	proc_context *ctx;
	int ret;

	ctx = __find_proc_context(pid);
	if (!ctx) {
		_E("Failed to find process(%d) context", pid);
		return -1;
	}

	if (ctx->watchdog_enable) {
		_W("Watchdog timer is already enabled. pid(%d)", pid);
		return 0;
	}

	if (!ctx->frozen)
		__watchdog_set_timer(ctx);

	ctx->watchdog_enable = true;

	ret = __send_request(ctx, WATCHDOG_ENABLE, __watchdog.interval);

	_I("pid(%d), result(%d)", pid, ret);
	_watchdog_logger_print("ENABLE", "pid(%d), appid(%s)",
			pid, __get_appid(pid));

	return 0;
}

static int __dispatch_watchdog_disable(amd_request_h req)
{
	int pid = amd_request_get_pid(req);
	proc_context *ctx;
	int ret;

	ctx = __find_proc_context(pid);
	if (!ctx) {
		_E("Failed to find process(%d) context", pid);
		return -1;
	}

	if (!ctx->watchdog_enable) {
		_W("Watchdog timer is already disabled. pid(%d)", pid);
		return 0;

	}

	__watchdog_unset_timer(ctx);

	ctx->watchdog_enable = false;
	ctx->retry_count = 0;
	ctx->kick_count = 0;
	ctx->resp_time.tv_sec = 0;
	ctx->resp_time.tv_nsec = 0;

	ret = __send_request(ctx, WATCHDOG_DISABLE, 0);

	_I("pid(%d), result(%d)", pid, ret);
	_watchdog_logger_print("DISABLE", "pid(%d), appid(%s)",
			pid, __get_appid(pid));

	return 0;
}

static int __dispatch_watchdog_kick(amd_request_h req)
{
	int pid = amd_request_get_pid(req);
	proc_context *ctx;

	ctx = __find_proc_context(pid);
	if (!ctx) {
		_E("Failed to find process(%d) context", pid);
		return -1;
	}

	if (!ctx->watchdog_enable) {
		_W("watchdog timer is not enabled. pid(%d)", pid);
		return -1;
	}

	__watchdog_unset_timer(ctx);
	__watchdog_set_timer(ctx);

	ctx->kick_count++;

	_I("pid(%d), kick count(%d)", pid, ctx->kick_count);
	_watchdog_logger_print("KICK", "count(%d), pid(%d), appid(%s)",
			ctx->kick_count, pid, __get_appid(pid));

	return 0;
}

static int __dispatch_watchdog_ping(amd_request_h req)
{
	int pid = amd_request_get_pid(req);
	proc_context *ctx;

	ctx = __find_proc_context(pid);
	if (!ctx) {
		_E("Failed to find process(%d) context", pid);
		return -1;
	}

	if (!ctx->watchdog_enable) {
		_W("watchdog timer is not enabled. pid(%d)", pid);
		return -1;
	}

	__watchdog_unset_timer(ctx);
	__watchdog_set_timer(ctx);
	clock_gettime(CLOCK_MONOTONIC, &ctx->resp_time);
	ctx->retry_count = 0;

	_I("pid(%d), resp_time(%ld.%ld)",
			pid, ctx->resp_time.tv_sec, ctx->resp_time.tv_nsec);
	_watchdog_logger_print("PING", "pid(%d), appid(%s)",
			pid, __get_appid(pid));

	return 0;
}

static int __on_app_status_cleanup(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	int pid = arg1;
	proc_context *ctx;

	ctx = __find_proc_context(pid);
	if (!ctx) {
		_W("Failed to find process(%d) context", pid);
		return AMD_NOTI_CONTINUE;
	}

	__watchdog.proc_contexts = g_list_remove(__watchdog.proc_contexts, ctx);
	__destroy_proc_context(ctx);

	return AMD_NOTI_CONTINUE;
}

static int __on_app_status_add(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	amd_app_status_h app_status = (amd_app_status_h)arg3;
	int pid = amd_app_status_get_pid(app_status);
	uid_t uid;
	proc_context *ctx;
	int operation_state;

	ctx = __find_proc_context(pid);
	if (ctx) {
		_W("Process(%d) context already exists", pid);
		return AMD_NOTI_CONTINUE;
	}

	uid = amd_app_status_get_uid(app_status);
	ctx = __create_proc_context(pid, uid);
	if (!ctx) {
		_E("Failed to create process(%d) context", pid);
		return AMD_NOTI_CONTINUE;
	}

	__watchdog.proc_contexts = g_list_append(__watchdog.proc_contexts, ctx);

	operation_state = _watchdog_config_get_operation_state();
	if (operation_state == WATCHDOG_ENABLE_BY_DEFAULT) {
		__watchdog_set_timer(ctx);
		ctx->watchdog_enable = true;
		_watchdog_logger_print("ENABLE", "pid(%d), appid(%s)",
				pid, amd_app_status_get_appid(app_status));
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_signal_send_watchdog_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	int pid = arg1;
	proc_context *ctx;

	_W("Watchdog pid(%d)", pid);
	ctx = __find_proc_context(pid);
	if (!ctx)
		return AMD_NOTI_CONTINUE;

	__watchdog_unset_timer(ctx);

	return AMD_NOTI_CONTINUE;
}

static amd_request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = WATCHDOG_ENABLE,
		.callback = __dispatch_watchdog_enable
	},
	{
		.cmd = WATCHDOG_DISABLE,
		.callback = __dispatch_watchdog_disable
	},
	{
		.cmd = WATCHDOG_KICK,
		.callback = __dispatch_watchdog_kick
	},
	{
		.cmd = WATCHDOG_PING,
		.callback = __dispatch_watchdog_ping
	},
};

static void __on_freezer_state_changed(GDBusConnection *connection,
		const gchar *sender_name,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *signal_name,
		GVariant *parameters,
		gpointer user_data)
{
	gint pid = -1;
	gint status = -1;
	proc_context *ctx;

	if (!g_strcmp0(signal_name, RESOURCED_FREEZER_SIGNAL)) {
		g_variant_get(parameters, "(ii)", &status, &pid);
		ctx = __find_proc_context(pid);
		if (!ctx) {
			_E("Failed to find process(%d) context", pid);
			return;
		}

		/* 0: SET_FOREGRD, 1: SET_BACKGRD, */
		ctx->frozen = (bool)status;
		if (ctx->frozen) {
			__watchdog_unset_timer(ctx);
		} else {
			if (ctx->watchdog_enable)
				__watchdog_set_timer(ctx);
		}

		_W("pid(%d), freezer state(%s)",
				pid, status ? "BACKGRD" : "FOREGRD");
		_watchdog_logger_print("FREEZER",
				"state(%s), pid(%d), appid(%s)",
				status ? "BACKGRD" : "FOREGRD",
				pid, __get_appid(pid));
	}
}

static int __listen_freezer_state(void *user_data)
{
	GError *err = NULL;

	if (__watchdog.listener)
		return 0;

	if (!__watchdog.conn) {
		__watchdog.conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
		if (!__watchdog.conn) {
			_E("Failed to connecto D-Bus daemon. err(%s)",
					err->message);
			g_error_free(err);
			return -1;
		}
	}

	__watchdog.listener = g_dbus_connection_signal_subscribe(
			__watchdog.conn,
			NULL,
			RESOURCED_FREEZER_INTERFACE,
			RESOURCED_FREEZER_SIGNAL,
			RESOURCED_FREEZER_PATH,
			NULL,
			G_DBUS_SIGNAL_FLAGS_NONE,
			__on_freezer_state_changed,
			NULL,
			NULL);
	if (!__watchdog.listener) {
		_E("Failed to subscribe freezer state");
		return -1;
	}

	_I("%s is subscribed", RESOURCED_FREEZER_SIGNAL);

	return 0;
}

static void __ignore_freezer_state(void)
{
	if (!__watchdog.conn)
		return;

	if (__watchdog.listener) {
		g_dbus_connection_signal_unsubscribe(__watchdog.conn,
				__watchdog.listener);
		__watchdog.listener = 0;
	}

	g_object_unref(__watchdog.conn);
	__watchdog.conn = NULL;
}

EXPORT int AMD_MOD_INIT(void)
{
	int r;

	_D("init");

	_watchdog_logger_init();
	_watchdog_config_init();

	__watchdog.interval = _watchdog_config_get_interval();
	__watchdog.max_retry_count = _watchdog_config_get_max_retry_count();

	r = amd_request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (r < 0) {
		_E("Failed to register cmds");
		return -1;
	}

	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_ADD,
			__on_app_status_add);
	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_CLEANUP,
			__on_app_status_cleanup);
	amd_noti_listen(AMD_NOTI_MSG_SIGNAL_SEND_WATCHDOG_START,
			__on_signal_send_watchdog_start);

	amd_signal_add_ready_cb(__listen_freezer_state, NULL);

	return 0;
}

EXPORT void AMD_MOD_FINI(void)
{
	_D("fini");

	if (__watchdog.retry_handler)
		g_source_remove(__watchdog.retry_handler);

	__ignore_freezer_state();

	if (__watchdog.proc_contexts) {
		g_list_free_full(__watchdog.proc_contexts,
				__destroy_proc_context);
	}

	_watchdog_config_fini();
	_watchdog_logger_fini();
}
