/*
 * Copyright (c) 2017 - 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <glib.h>
#include <gio/gio.h>
#include <wayland-client.h>
#include <wayland-tbm-client.h>
#include <tizen-extension-client-protocol.h>
#include <amd.h>

#include "wayland-core-private.h"

#define PATH_RUN_WAYLAND "/run/wayland-0"
#define PATH_RUN_WMREADY "/run/.wm_ready"
#define PATH_RUN "/run"

struct file_monitor_s {
	bool wl_ready;
	bool wm_ready;
	amd_inotify_watch_info_h handle;
};

struct wl_source_s {
	struct wl_display *display;
	struct wl_registry *registry;
	GSource *gsource;
	GPollFD *gfd;
	bool prepared;
};

static struct file_monitor_s __file_monitor;
static struct wl_source_s __wl_source;
static int __retry_cnt;

static void __wl_listener_cb(void *data, struct wl_registry *reg,
		unsigned int id, const char *interface, unsigned int version)
{
	char buf[512];

	if (!interface)
		return;

	LOGW("interface(%s), id(%u), version(%u)", interface, id, version);
	snprintf(buf, sizeof(buf), "wayland.listener.%s", interface);
	amd_noti_send(buf, (int)id, (int)version, reg, NULL);
}

static void __wl_listener_remove_cb(void *data, struct wl_registry *reg,
		unsigned int id)
{
	LOGW("id(%u)", id);
	amd_noti_send(AMD_NOTI_MSG_WAYLAND_LISTENER_REMOVE,
			(int)id, 0, reg, NULL);
}

static const struct wl_registry_listener __registry_listener = {
	__wl_listener_cb,
	__wl_listener_remove_cb
};

static gboolean __wl_source_prepare(GSource *source, gint *timeout)
{
	if (__wl_source.prepared)
		return FALSE;

	while (wl_display_prepare_read(__wl_source.display) != 0) {
		if (wl_display_dispatch_pending(__wl_source.display) < 0)
			LOGW("Failed to dispatch queue pending");
	}

	__wl_source.prepared = true;

	wl_display_flush(__wl_source.display);
	*timeout = -1;

	return FALSE;
}

static gboolean __wl_source_check(GSource *source)
{
	if (g_source_is_destroyed(source)) {
		if (__wl_source.display && __wl_source.prepared)
			wl_display_cancel_read(__wl_source.display);
		LOGE("Display source(%p) already destroyed", source);
		return FALSE;
	}

	if (!__wl_source.prepared)
		return FALSE;

	if (__wl_source.gfd->revents & (G_IO_HUP | G_IO_ERR)) {
		wl_display_cancel_read(__wl_source.display);
		LOGE("Error: revents(%x)", __wl_source.gfd->revents);
		__wl_source.prepared = false;
		g_source_destroy(source);
		abort();
		return FALSE;
	}

	if (__wl_source.gfd->revents & G_IO_IN) {
		if (wl_display_read_events(__wl_source.display) < 0)
			LOGW("Failed to read events");
		__wl_source.prepared = false;
		return TRUE;
	}

	wl_display_cancel_read(__wl_source.display);
	__wl_source.prepared = false;

	return FALSE;
}

static gboolean __wl_source_dispatch(GSource *source, GSourceFunc callback,
		gpointer data)
{
	if (g_source_is_destroyed(source)) {
		LOGE("Source(%p) is already destroyed", source);
		return G_SOURCE_REMOVE;
	}

	if (__wl_source.gfd->revents & (G_IO_HUP | G_IO_ERR)) {
		LOGE("Error: revents(%x)", __wl_source.gfd->revents);
		g_source_destroy(source);
		abort();
		return G_SOURCE_REMOVE;
	}

	if (__wl_source.gfd->revents & G_IO_IN) {
		if (wl_display_dispatch_pending(__wl_source.display) < 0)
			LOGW("Failed to dispatch queue pending");
	}

	wl_display_flush(__wl_source.display);

	return G_SOURCE_CONTINUE;
}

static GSourceFuncs __wl_source_funcs = {
	.prepare = __wl_source_prepare,
	.check = __wl_source_check,
	.dispatch = __wl_source_dispatch,
	.finalize = NULL
};

static void __wl_source_fini(void)
{
	if (__wl_source.gfd) {
		g_source_remove_poll(__wl_source.gsource, __wl_source.gfd);
		g_free(__wl_source.gfd);
		__wl_source.gfd = NULL;
	}

	if (__wl_source.gsource != NULL &&
			!g_source_is_destroyed(__wl_source.gsource)) {
		g_source_destroy(__wl_source.gsource);
		g_source_unref(__wl_source.gsource);
		__wl_source.gsource = NULL;
	}

	if (__wl_source.registry) {
		wl_registry_destroy(__wl_source.registry);
		__wl_source.registry = NULL;
	}

	if (__wl_source.display) {
		wl_display_disconnect(__wl_source.display);
		__wl_source.display = NULL;
		amd_wayland_set_display(NULL);
	}
}

static int __wl_source_init(void)
{
	int r;

	__wl_source.display = wl_display_connect(NULL);
	if (!__wl_source.display) {
		LOGE("Failed to connect wayland display");
		return -1;
	}
	amd_wayland_set_display(__wl_source.display);

	__wl_source.registry = wl_display_get_registry(__wl_source.display);
	if (!__wl_source.registry) {
		LOGE("Failed to get wayland registry");
		return -1;
	}

	wl_registry_add_listener(__wl_source.registry,
			&__registry_listener, NULL);
	wl_display_flush(__wl_source.display);

	r = wl_display_roundtrip(__wl_source.display);
	if (r < 0)
		LOGW("wl_display_roundtrip() is failed. result(%d)", r);

	__wl_source.gsource = g_source_new(&__wl_source_funcs, sizeof(GSource));
	if (!__wl_source.gsource) {
		LOGE("Failed to create GSource");
		return -1;
	}

	__wl_source.gfd = (GPollFD *)g_malloc(sizeof(GPollFD));
	if (!__wl_source.gfd) {
		LOGE("Failed to create GPollFD");
		return -1;
	}

	__wl_source.gfd->fd = wl_display_get_fd(__wl_source.display);
	__wl_source.gfd->events = G_IO_IN | G_IO_ERR;
	__wl_source.gfd->revents = 0;

	g_source_add_poll(__wl_source.gsource, __wl_source.gfd);
	g_source_set_callback(__wl_source.gsource, NULL,
			__wl_source.display, NULL);
	g_source_set_priority(__wl_source.gsource, G_PRIORITY_DEFAULT);
	g_source_attach(__wl_source.gsource, NULL);

	return 0;
}

static gboolean __init_wl(gpointer data)
{
	LOGI("Initialize Wayland");
	if (__wl_source_init() < 0) {
		__retry_cnt++;
		if (__retry_cnt > 10) {
			LOGE("Failed to initialize wayland: Retry(%d)", __retry_cnt);
			__wl_source_fini();
			g_timeout_add_seconds(1, __init_wl, NULL);
			return G_SOURCE_REMOVE;
		}
		__wl_source_fini();
		return G_SOURCE_CONTINUE;
	}

	return G_SOURCE_REMOVE;
}

static bool __wayland_monitor_cb(const char *event_name, void *data)
{
	if (event_name == NULL)
		return true;

	if (strcmp(event_name, "wayland-0") == 0) {
		LOGD("%s is created", event_name);
		__file_monitor.wl_ready = true;
	} else if (strcmp(event_name, ".wm_ready") == 0) {
		LOGD("%s is created", event_name);
		__file_monitor.wm_ready = true;
	}

	if (__file_monitor.wm_ready && __file_monitor.wl_ready) {
		g_idle_add(__init_wl, NULL);
		__file_monitor.handle = NULL;
		return false;
	}

	return true;
}

static int __file_monitor_init(void)
{
	if (!__file_monitor.wl_ready) {
		if (access(PATH_RUN_WAYLAND, F_OK) == 0) {
			LOGD("%s exists", PATH_RUN_WAYLAND);
			__file_monitor.wl_ready = true;
		}
	}

	if (!__file_monitor.wm_ready) {
		if (access(PATH_RUN_WMREADY, F_OK) == 0) {
			LOGD("%s exists", PATH_RUN_WMREADY);
			__file_monitor.wm_ready = true;
		}
	}

	if (__file_monitor.wl_ready && __file_monitor.wm_ready) {
		g_idle_add(__init_wl, NULL);
		return 0;
	}

	__file_monitor.handle = amd_inotify_add_watch(PATH_RUN, IN_CREATE,
			__wayland_monitor_cb, NULL);
	if (__file_monitor.handle == NULL) {
		LOGE("Failed to add inotify watch");
		return -1;
	}

	return 0;
}

static void __file_monitor_fini(void)
{
	if (__file_monitor.handle)
		amd_inotify_rm_watch(__file_monitor.handle);
}

static gboolean __idle_cb(gpointer data)
{
	LOGD("wayland core init");

	if (__file_monitor_init() < 0)
		return G_SOURCE_CONTINUE;

	return G_SOURCE_REMOVE;
}

EXPORT int AMD_MOD_INIT(void)
{
	LOGD("wayland core init");

	g_idle_add(__idle_cb, NULL);

	return 0;
}

EXPORT void AMD_MOD_FINI(void)
{
	LOGD("wayland core finish");

	__file_monitor_fini();
	__wl_source_fini();
	__retry_cnt = 0;
}
