/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <glib.h>
#include <aul.h>
#include <aul_cmd.h>
#include <bundle_internal.h>
#include <cert-svc/ccert.h>
#include <cert-svc/cinstance.h>

#include "amd_config.h"
#include "amd_util.h"
#include "amd_signal.h"
#include "amd_anr_monitor.h"
#include "amd_app_status.h"
#include "amd_app_property.h"
#include "amd_appinfo.h"
#include "amd_noti.h"
#include "amd_request.h"
#include "amd_launch.h"

#define MAX_TIMEOUT 5000
#define METADATA_ANR_CHECK_BYPASS \
	"http://tizen.org/metadata/appmanager/anr_check_bypass"

typedef struct proc_info_s {
	pid_t pid;
	int ref;
	guint timer;
} proc_info_t;

static GHashTable *__proc_tbl;

static proc_info_t *__find_proc_info(pid_t pid);
static void __remove_proc_info(pid_t pid);

static bool __is_managed_by_taskmgr(app_status_h app_status)
{
	struct appinfo *ai;
	const char *taskmanage;
	const char *appid;
	uid_t uid;

	appid = _app_status_get_appid(app_status);
	uid = _app_status_get_uid(app_status);
	ai = _appinfo_find(uid, appid);
	if (!ai)
		return false;

	taskmanage = _appinfo_get_value(ai, AIT_TASKMANAGE);
	if (!strcmp(taskmanage, "true"))
		return true;

	return false;
}

static bool __can_ignore_anr_policy(app_status_h app_status)
{
	app_property_h app_property;
	const char *appid;
	bool ignore;
	uid_t uid;

	appid = _app_status_get_appid(app_status);
	uid = _app_status_get_uid(app_status);
	if (!_appinfo_is_platform_app(appid, uid))
		return false;

	app_property = _app_property_find(uid);
	ignore = _app_property_metadata_match(app_property,
			appid, METADATA_ANR_CHECK_BYPASS, "yes");
	if (ignore)
		return true;

	return false;
}

static gboolean __timeout_handler(gpointer data)
{
	pid_t pid = GPOINTER_TO_INT(data);
	app_status_h app_status;
	proc_info_t *ctx;

	ctx = __find_proc_info(pid);
	if (!ctx)
		return G_SOURCE_REMOVE;

	ctx->timer = 0;

	app_status = _app_status_find_v2(pid);
	if (!app_status) {
		_W("Failed to find app status. pid(%d)", pid);
		__remove_proc_info(pid);
		return G_SOURCE_REMOVE;
	}

	if (_app_status_get_status(app_status) == STATUS_DYING) {
		_W("%d is dying", pid);
		__remove_proc_info(pid);
		return G_SOURCE_REMOVE;
	}

	_W("Application(%d) Not Responding", pid);
	if (__can_ignore_anr_policy(app_status)) {
		_W("Ignore ANR policy. pid(%d)", pid);
		return G_SOURCE_REMOVE;
	}

	if (!__is_managed_by_taskmgr(app_status)) {
		_W("The process(%d) is not managed by task-manager", pid);
		return G_SOURCE_REMOVE;
	}

	__remove_proc_info(pid);
	_signal_send_watchdog(pid, SIGKILL);

	return G_SOURCE_REMOVE;
}

static proc_info_t *__create_proc_info(pid_t pid)
{
	proc_info_t *ctx;

	ctx = calloc(1, sizeof(proc_info_t));
	if (!ctx) {
		_E("Out of memory");
		return NULL;
	}

	ctx->pid = pid;
	ctx->ref = 1;

	ctx->timer = g_timeout_add(MAX_TIMEOUT, __timeout_handler,
			GINT_TO_POINTER(pid));
	if (ctx->timer == 0)
		_E("Failed to add timer");

	return ctx;
}

static void __destroy_proc_info(gpointer data)
{
	proc_info_t *ctx = (proc_info_t *)data;

	if (!ctx) {
		_E("Critical error");
		return;
	}

	if (ctx->timer > 0)
		g_source_remove(ctx->timer);

	free(ctx);
}

static proc_info_t *__find_proc_info(pid_t pid)
{
	if (!g_hash_table_contains(__proc_tbl, GINT_TO_POINTER(pid)))
		return NULL;

	return g_hash_table_lookup(__proc_tbl, GINT_TO_POINTER(pid));
}

static void __add_proc_info(proc_info_t *ctx)
{
	if (g_hash_table_contains(__proc_tbl, GINT_TO_POINTER(ctx->pid)))
		return;

	g_hash_table_insert(__proc_tbl, GINT_TO_POINTER(ctx->pid), ctx);
}

static void __remove_proc_info(pid_t pid)
{
	if (!g_hash_table_contains(__proc_tbl, GINT_TO_POINTER(pid)))
		return;

	g_hash_table_remove(__proc_tbl, GINT_TO_POINTER(pid));
}

static void __reset_timer(proc_info_t *ctx)
{
	if (ctx->timer > 0)
		g_source_remove(ctx->timer);

	ctx->timer = g_timeout_add(MAX_TIMEOUT, __timeout_handler,
			GINT_TO_POINTER(ctx->pid));
	if (ctx->timer == 0)
		_E("Failed to add timer. pid(%d)", ctx->pid);

	_W("Reset timer. pid(%d)", ctx->pid);
}

int _anr_monitor_add_timer(pid_t pid)
{
	proc_info_t *ctx;

	_W("Add timer. pid(%d)", pid);
	ctx = __find_proc_info(pid);
	if (ctx) {
		__reset_timer(ctx);
		ctx->ref++;
		return 0;
	}

	ctx = __create_proc_info(pid);
	if (!ctx) {
		_E("Failed to create process(%d) info", pid);
		return -1;
	}

	__add_proc_info(ctx);

	return 0;
}

int _anr_monitor_remove_timer(pid_t pid)
{
	proc_info_t *ctx;

	_W("Remove timer. pid(%d)", pid);
	ctx = __find_proc_info(pid);
	if (!ctx) {
		_E("Failed to find process(%d) info", pid);
		return -1;
	}

	ctx->ref--;
	if (ctx->ref > 0)
		return 0;

	__remove_proc_info(pid);

	return 0;
}

static int __dispatch_anr_notify(request_h req)
{
	bundle *b = _request_get_bundle(req);
	uid_t uid = _request_get_uid(req);
	int pid = _request_get_pid(req);
	app_status_h app_status;
	const char *cmd_str;
	int cmd;

	cmd_str = bundle_get_val(b, AUL_K_COMMAND);
	if (cmd_str)
		cmd = atoi(cmd_str);
	else
		cmd = -1;

	_E("Application(%d) Not Responding. cmd(%d)", pid, cmd);

	app_status = _app_status_find_v2(pid);
	if (!app_status) {
		_E("Failed to find app status. pid(%d)", pid);
		return -1;
	}

	switch (cmd) {
	case APP_OPEN:
	case APP_RESUME:
	case APP_START:
	case APP_START_RES:
	case APP_START_ASYNC:
	case APP_START_RES_ASYNC:
	case APP_SEND_LAUNCH_REQUEST:
	case APP_SEND_LAUNCH_REQUEST_SYNC:
		if (!__can_ignore_anr_policy(app_status) &&
				__is_managed_by_taskmgr(app_status))
			_signal_send_watchdog(pid, SIGKILL);
		break;
	case APP_TERM_BY_PID:
	case APP_TERM_REQ_BY_PID:
	case APP_TERM_BY_PID_ASYNC:
	case APP_TERM_BGAPP_BY_PID:
	case APP_TERM_INSTANCE_ASYNC:
	case APP_TERM_BG_INSTANCE:
		_launch_send_sigkill(pid, uid);
		break;
	default:
		break;
	}

	return 0;
}

static request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = ANR_NOTIFY,
		.callback = __dispatch_anr_notify
	},
};

static int __on_signal_send_watchdog_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int pid = arg1;

	_W("Watchdog pid(%d)", pid);
	_anr_monitor_remove_timer(pid);

	return 0;
}

int _anr_monitor_init(void)
{
	int ret;

	_D("[__ANR_MONITOR__] init");

	__proc_tbl = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, __destroy_proc_info);
	if (!__proc_tbl) {
		_E("Failed to create proc table");
		return -1;
	}

	ret = _app_property_metadata_add_filter(METADATA_ANR_CHECK_BYPASS,
			NULL);
	if (ret < 0)
		return -1;

	ret = _request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (ret < 0)
		return -1;

	_noti_listen(NOTI_MSG_SIGNAL_SEND_WATCHDOG_START,
			__on_signal_send_watchdog_start);

	return 0;
}

void _anr_monitor_fini(void)
{
	_D("[__ANR_MONITOR__] fini");

	_app_property_metadata_remove_filter(METADATA_ANR_CHECK_BYPASS, NULL);

	if (__proc_tbl) {
		g_hash_table_destroy(__proc_tbl);
		__proc_tbl = NULL;
	}
}
