/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <uuid/uuid.h>

#include <glib.h>
#include <aul.h>
#include <aul_cmd.h>
#include <aul_sock.h>
#include <bundle_internal.h>

#include "amd_config.h"
#include "amd_util.h"
#include "amd_request.h"
#include "amd_noti.h"
#include "amd_compinfo.h"
#include "amd_comp_status.h"
#include "amd_socket.h"
#include "amd_app_status.h"
#include "amd_launch.h"
#include "amd_cynara.h"

struct comp_status_s {
	char *comp_id;
	char *instance_id;
	char *leader_instance_id;
	pid_t pid;
	int comp_type;
	int window;
	int status;
	bool is_subcomp;
	bool started;
};

static GList *__comp_list;

const char *_comp_status_generate_instance(const char *comp_id)
{
	static char buf[MAX_PACKAGE_STR_SIZE];
	char uuid[37];
	uuid_t u;

	if (!comp_id) {
		_E("Invalid parameter");
		return NULL;
	}

	uuid_generate(u);
	uuid_unparse(u, uuid);
	snprintf(buf, sizeof(buf), "%s:%s", uuid, comp_id);
	return buf;
}

static void __destroy_comp_status(gpointer data)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)data;

	if (!comp_status)
		return;

	if (comp_status->instance_id)
		free(comp_status->instance_id);
	if (comp_status->comp_id)
		free(comp_status->comp_id);
	free(comp_status);
}

static int __convert_comp_type(const char *type)
{
	if (!type)
		return -1;

	if (!strcmp(type, "frame"))
		return CT_FRAME_COMP;
	else if (!strcmp(type, "service"))
		return CT_SERVICE_COMP;

	return -1;
}

static const char *__get_comp_type_string(int comp_type)
{
	switch (comp_type) {
	case CT_FRAME_COMP:
		return "frame";
	case CT_SERVICE_COMP:
		return "service";
	default:
		return "Unknown";
	}
}

static struct comp_status_s *__create_comp_status(compinfo_h ci,
		const char *instance_id, pid_t pid, bool is_subcomp)
{
	struct comp_status_s *comp_status;
	const char *comp_id;
	const char *type;

	comp_status = calloc(1, sizeof(struct comp_status_s));
	if (!comp_status) {
		_E("Out of memory");
		return NULL;
	}

	comp_id = _compinfo_get_value(ci, CIT_NAME);
	if (!comp_id) {
		_E("Failed to get component ID");
		free(comp_status);
		return NULL;
	}

	comp_status->comp_id = strdup(comp_id);
	if (!comp_status->comp_id) {
		_E("Failed to duplicate component ID(%s)", comp_id);
		__destroy_comp_status(comp_status);
		return NULL;
	}

	comp_status->instance_id = strdup(instance_id);
	if (!comp_status->instance_id) {
		_E("Failed to duplicate instance ID(%s)", instance_id);
		__destroy_comp_status(comp_status);
		return NULL;
	}

	comp_status->pid = pid;
	comp_status->is_subcomp = is_subcomp;
	comp_status->started = false;

	type = _compinfo_get_value(ci, CIT_TYPE);
	comp_status->comp_type = __convert_comp_type(type);
	comp_status->status = COMP_STATUS_INITIALIZED;

	return comp_status;
}

static gint __compare_window(gconstpointer a, gconstpointer b)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)a;
	int window = GPOINTER_TO_INT(b);

	return (comp_status->window != window);
}

static gint __compare_instance(gconstpointer a, gconstpointer b)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)a;
	const char *instance_id = (const char *)b;

	return strcmp(comp_status->instance_id, instance_id);
}

int _comp_status_add_comp_info(compinfo_h ci, pid_t pid,
		const char *instance_id, bool is_subcomp)
{
	struct comp_status_s *comp_status;
	GList *found;

	if (!ci || pid < 0 || !instance_id)
		return -1;

	found = g_list_find_custom(__comp_list, instance_id,
			__compare_instance);
	if (found) {
		_D("Already exists. %d:%s", pid, instance_id);
		return 0;
	}

	comp_status = __create_comp_status(ci, instance_id, pid, is_subcomp);
	if (!comp_status) {
		_E("Failed to create component status. %d:%s",
				pid, instance_id);
		return -1;
	}

	__comp_list = g_list_append(__comp_list, comp_status);

	_noti_send(NOTI_MSG_COMP_STATUS_ADD, 0, 0, (void *)comp_status, NULL);

	return 0;
}

comp_status_h _comp_status_find(const char *comp_id)
{
	struct comp_status_s *comp_status;
	GList *iter;

	if (!comp_id) {
		_E("Invalid parameter");
		return NULL;
	}

	iter = __comp_list;
	while (iter) {
		comp_status = (struct comp_status_s *)iter->data;
		if (!comp_status->is_subcomp &&
				!strcmp(comp_status->comp_id, comp_id))
			return comp_status;

		iter = g_list_next(iter);
	}

	return NULL;
}

comp_status_h _comp_status_find_by_instance_id(const char *instance_id)
{
	GList *found;

	if (!instance_id) {
		_E("Invalid parameter");
		return NULL;
	}

	found = g_list_find_custom(__comp_list, instance_id,
			__compare_instance);
	if (!found)
		return NULL;

	return found->data;
}

comp_status_h _comp_status_find_by_window(int window)
{
	GList *found;

	if (window <= 0) {
		_E("Invalid parameter");
		return NULL;
	}

	found = g_list_find_custom(__comp_list, GINT_TO_POINTER(window),
			__compare_window);
	if (!found)
		return NULL;

	return found->data;
}

comp_status_h _comp_status_find_v2(pid_t pid, const char *instance_id)
{
	struct comp_status_s *comp_status;
	pid_t pgid;

	comp_status = _comp_status_find_by_instance_id(instance_id);
	if (!comp_status)
		return NULL;

	if (comp_status->pid != pid) {
		pgid = getpgid(pid);
		if (comp_status->pid == pgid) {
			_E("Process ID(%d:%d) is not equal to %d",
					pid, pgid, comp_status->pid);
			return NULL;
		}
	}

	return comp_status;
}

pid_t _comp_status_get_pid(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return -1;

	return comp_status->pid;
}

const char *_comp_status_get_comp_id(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return NULL;

	return comp_status->comp_id;
}

const char *_comp_status_get_instance_id(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return NULL;

	return comp_status->instance_id;
}

int _comp_status_set_leader_id(comp_status_h h, const char *instance_id)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status || instance_id == NULL)
		return -1;

	if (comp_status->leader_instance_id) {
		free(comp_status->leader_instance_id);
		comp_status->leader_instance_id = NULL;
	}

	comp_status->leader_instance_id = strdup(instance_id);
	if (comp_status->leader_instance_id == NULL) {
		_E("Fail to dup id");
		return -1;
	}

	return 0;
}

int _comp_status_set_window(comp_status_h h, int window)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return -1;

	if (window <= 0)
		return -1;

	comp_status->window = window;

	return 0;
}

const char *_comp_status_get_leader_id(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return NULL;

	return comp_status->leader_instance_id;
}

int _comp_status_get_comp_type(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return -1;

	return comp_status->comp_type;
}

int _comp_status_get_status(comp_status_h h)
{
	struct comp_status_s *comp_status = (struct comp_status_s *)h;

	if (!comp_status)
		return -1;

	return comp_status->status;
}

static int __on_main_app_dead(const char *msg, int arg1, int arg2, void *arg3,
		bundle *b)
{
	struct comp_status_s *comp_status;
	pid_t pid = (pid_t)arg1;
	GList *iter;

	iter = __comp_list;
	while (iter) {
		comp_status = (struct comp_status_s *)iter->data;
		iter = g_list_next(iter);
		if (comp_status->pid == pid) {
			__comp_list = g_list_remove(__comp_list, comp_status);
			__destroy_comp_status(comp_status);
		}
	}

	return NOTI_CONTINUE;
}

static int __dispatch_comp_notify_start(request_h req)
{
	struct comp_status_s *comp_status;
	pid_t pid = _request_get_pid(req);
	bundle *b = _request_get_bundle(req);
	const char *instance_id;

	if (!b) {
		_E("Failed to get bundle. pid(%d)", pid);
		return -1;
	}

	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	comp_status = _comp_status_find_v2(pid, instance_id);
	if (!comp_status) {
		_E("Failed to find component status. pid(%d)", pid);
		return -1;
	}

	comp_status->started = true;

	_W("[COMP_NOTIFY_START] pid(%d), instance_id(%s)", pid, instance_id);

	return 0;
}

static int __dispatch_comp_notify_exit(request_h req)
{
	struct comp_status_s *comp_status;
	pid_t pid = _request_get_pid(req);
	uid_t uid = _request_get_uid(req);
	bundle *b = _request_get_bundle(req);
	const char *instance_id;

	if (!b) {
		_E("Failed to get bundle. pid(%d)", pid);
		return -1;
	}

	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	comp_status = _comp_status_find_v2(pid, instance_id);
	if (!comp_status) {
		_E("Failed to find component status. pid(%d)", pid);
		return -1;
	}

	__comp_list = g_list_remove(__comp_list, comp_status);
	__destroy_comp_status(comp_status);

	_noti_send(NOTI_MSG_COMP_STATUS_NOTIFY_EXIT, (int)pid, (int)uid,
			(void *)instance_id, NULL);

	_W("[COMP_NOTIFY_EXIT] pid(%d), instance_id(%s)", pid, instance_id);

	return 0;
}

static int __dispatch_comp_status_update(request_h req)
{
	struct comp_status_s *comp_status;
	pid_t pid = _request_get_pid(req);
	bundle *b = _request_get_bundle(req);
	const char *instance_id;
	const char *status_str;
	bool update_group_info = false;
	int status;

	if (!b) {
		_E("Failed to get bundle. pid(%d)", pid);
		return -1;
	}

	status_str = bundle_get_val(b, AUL_K_STATUS);
	if (!status_str) {
		_E("Invalid request. pid(%d)", pid);
		return -1;
	}

	status = atoi(status_str);

	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	comp_status = _comp_status_find_v2(pid, instance_id);
	if (!comp_status) {
		_E("Failed to find component status. pid(%d)", pid);
		return -1;
	}

	comp_status->status = status;

	if (comp_status->comp_type == CT_FRAME_COMP)
		update_group_info = true;

	_noti_send(NOTI_MSG_COMP_STATUS_UPDATE_STATUS_END,
			pid, update_group_info, comp_status, NULL);

	_W("[COMP_STATUS_UPDATE] pid(%d), instance(%s), status(%d)",
			pid, instance_id, comp_status->status);

	return 0;
}

static bundle *__create_compinfo_bundle(struct comp_status_s *comp_status)
{
	char buf[MAX_PID_STR_BUFSZ];
	const char *comp_type;
	const char *appid = NULL;
	app_status_h app_status;
	bundle *b;

	b = bundle_create();
	if (!b)
		return NULL;

	app_status = _app_status_find_v2(comp_status->pid);
	if (app_status)
		appid = _app_status_get_appid(app_status);

	if (appid)
		bundle_add(b, AUL_K_APPID, appid);

	bundle_add(b, AUL_K_COMPONENT_ID, comp_status->comp_id);
	bundle_add(b, AUL_K_INSTANCE_ID, comp_status->instance_id);

	snprintf(buf, sizeof(buf), "%d", comp_status->pid);
	bundle_add(b, AUL_K_PID, buf);

	comp_type = __get_comp_type_string(comp_status->comp_type);
	bundle_add(b, AUL_K_COMPONENT_TYPE, comp_type);

	snprintf(buf, sizeof(buf), "%d", comp_status->status);
	bundle_add(b, AUL_K_STATUS, buf);

	snprintf(buf, sizeof(buf), "%d", comp_status->is_subcomp);
	bundle_add(b, AUL_K_IS_SUB_COMP, buf);

	return b;
}

static int __send_running_compinfo(struct comp_status_s *comp_status, int fd)
{
	bundle *b;
	int ret;

	b = __create_compinfo_bundle(comp_status);
	if (!b) {
		_E("Out of memory");
		aul_sock_send_raw_with_fd(fd, APP_GET_INFO_ERROR,
				NULL, 0, AUL_SOCK_ASYNC);
		return -1;
	}

	ret = aul_sock_send_bundle_with_fd(fd, APP_GET_INFO_OK, b,
			AUL_SOCK_ASYNC);
	if (ret < 0) {
		_E("Failed to send bundle data. error(%d)", ret);
		bundle_free(b);
		return -1;
	}
	bundle_free(b);

	return 0;
}

static int __dispatch_comp_context_foreach(request_h req)
{
	struct comp_status_s *comp_status;
	int fd = _request_remove_fd(req);
	GList *iter;
	int count;
	int ret;

	count = g_list_length(__comp_list);
	if (count == 0) {
		_E("Components are not running");
		_send_result_to_client(fd, -ENOENT);
		return -1;
	}
	_send_result_to_client_v2(fd, count);

	for (iter = __comp_list; iter; iter = g_list_next(iter)) {
		comp_status = (struct comp_status_s *)iter->data;
		if (!comp_status)
			continue;

		ret = __send_running_compinfo(comp_status, fd);
		if (ret < 0) {
			close(fd);
			return -1;
		}
	}
	close(fd);

	return 0;
}

static int __dispatch_comp_context_get(request_h req)
{
	struct comp_status_s *comp_status;
	bundle *b = _request_get_bundle(req);
	int fd = _request_remove_fd(req);
	const char *comp_id;
	int ret;

	if (!b) {
		_E("Invalid parameter");
		aul_sock_send_raw_with_fd(fd, AUL_R_EINVAL, NULL, 0,
				AUL_SOCK_NOREPLY);
		return -1;
	}

	comp_id = bundle_get_val(b, AUL_K_COMPONENT_ID);
	if (!comp_id) {
		_E("Failed to get component ID");
		aul_sock_send_raw_with_fd(fd, AUL_R_EINVAL, NULL, 0,
				AUL_SOCK_NOREPLY);
		return -1;
	}

	comp_status = _comp_status_find(comp_id);
	if (!comp_status) {
		_E("Failed to find running component(%s) context",
				comp_id);
		aul_sock_send_raw_with_fd(fd, AUL_R_ENOENT, NULL, 0,
				AUL_SOCK_NOREPLY);
		return -1;
	}

	ret = __send_running_compinfo(comp_status, fd);
	if (ret < 0) {
		close(fd);
		return -1;
	}
	close(fd);

	return 0;
}

static int __dispatch_comp_context_is_running(request_h req)
{
	struct comp_status_s *comp_status;
	bundle *b = _request_get_bundle(req);
	const char *instance_id;
	int is_running;

	if (!b) {
		_E("Invalid parameter");
		_request_send_result(req, -EINVAL);
		return -1;
	}

	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!instance_id) {
		_E("Failed to get instance ID");
		_request_send_result(req, -EINVAL);
		return -1;
	}

	comp_status = _comp_status_find_by_instance_id(instance_id);
	if (!comp_status || comp_status->status == COMP_STATUS_DESTROYED)
		is_running = 0;
	else
		is_running = 1;

	_request_send_result(req, is_running);
	_I("[COMP_CONTEXT_IS_RUNNING] is_running(%s)",
			is_running ? "true" : "false");

	return 0;
}

static int __comp_context_process(request_h req)
{
	struct comp_status_s *comp_status;
	bundle *b = _request_get_bundle(req);
	const char *instance_id;
	int cmd;
	int ret;

	if (!b) {
		_E("Invalid parameter");
		_request_send_result(req, -EINVAL);
		return -1;
	}

	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!instance_id) {
		_E("Failed to get instance ID");
		_request_send_result(req, -EINVAL);
		return -1;
	}

	comp_status = _comp_status_find_by_instance_id(instance_id);
	if (!comp_status) {
		_E("Failed to find running context. instance(%s)", instance_id);
		_request_send_result(req, -ENOENT);
		return -1;
	}

	cmd = _request_get_cmd(req);
	switch (cmd) {
	case COMP_CONTEXT_RESUME:
	case COMP_CONTEXT_PAUSE:
		aul_send_app_resume_request_signal(comp_status->pid, NULL,
				NULL, NULL);
		break;
	case COMP_CONTEXT_TERMINATE_BG_COMP:
	case COMP_CONTEXT_TERMINATE:
		aul_send_app_terminate_request_signal(comp_status->pid, NULL,
				NULL, NULL);
		break;
	default:
		break;
	}

	switch (cmd) {
	case COMP_CONTEXT_RESUME:
		ret = _launch_resume_inst(comp_status->pid, req);
		break;
	case COMP_CONTEXT_PAUSE:
		ret = _launch_pause_inst(comp_status->pid, req);
		break;
	case COMP_CONTEXT_TERMINATE_BG_COMP:
		ret = _launch_terminate_bg_inst(comp_status->pid, req);
		break;
	case COMP_CONTEXT_TERMINATE:
		ret = _launch_terminate_inst(comp_status->pid, req);
		break;
	default:
		_E("Unknown command: %d", cmd);
		ret = -1;
		break;
	}

	return ret;
}

static int __dispatch_comp_context_resume(request_h req)
{
	return __comp_context_process(req);
}

static int __dispatch_comp_context_pause(request_h req)
{
	return __comp_context_process(req);
}

static int __dispatch_comp_context_terminate_bg_comp(request_h req)
{
	int fd = _request_remove_fd(req);
	int ret;

	ret = __comp_context_process(req);
	_send_result_to_client(fd, ret);

	return ret;
}

static int __dispatch_comp_context_terminate(request_h req)
{
	int fd = _request_remove_fd(req);
	int ret;

	ret = __comp_context_process(req);
	_send_result_to_client(fd, ret);

	return ret;
}

static request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = COMP_NOTIFY_START,
		.callback = __dispatch_comp_notify_start
	},
	{
		.cmd = COMP_NOTIFY_EXIT,
		.callback = __dispatch_comp_notify_exit
	},
	{
		.cmd = COMP_STATUS_UPDATE,
		.callback = __dispatch_comp_status_update
	},
	{
		.cmd = COMP_CONTEXT_FOREACH,
		.callback = __dispatch_comp_context_foreach
	},
	{
		.cmd = COMP_CONTEXT_GET,
		.callback = __dispatch_comp_context_get
	},
	{
		.cmd = COMP_CONTEXT_IS_RUNNING,
		.callback = __dispatch_comp_context_is_running
	},
	{
		.cmd = COMP_CONTEXT_RESUME,
		.callback = __dispatch_comp_context_resume
	},
	{
		.cmd = COMP_CONTEXT_PAUSE,
		.callback = __dispatch_comp_context_pause
	},
	{
		.cmd = COMP_CONTEXT_TERMINATE_BG_COMP,
		.callback = __dispatch_comp_context_terminate_bg_comp
	},
	{
		.cmd = COMP_CONTEXT_TERMINATE,
		.callback = __dispatch_comp_context_terminate
	},
};

static cynara_checker __cynara_checkers[] = {
	{
		.cmd = COMP_CONTEXT_RESUME,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_APPMANAGER_LAUNCH
	},
	{
		.cmd = COMP_CONTEXT_PAUSE,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_APPMANAGER_LAUNCH
	},
	{
		.cmd = COMP_CONTEXT_TERMINATE_BG_COMP,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_APPMANAGER_KILL_BGAPP
	},
	{
		.cmd = COMP_CONTEXT_TERMINATE,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_APPMANAGER_KILL
	},
	{
		.cmd = COMP_CONTEXT_GET,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_PACKAGEMANAGER_INFO
	},
	{
		.cmd = COMP_CONTEXT_FOREACH,
		.checker = _cynara_simple_checker,
		.data = PRIVILEGE_PACKAGEMANAGER_INFO
	}
};

int _comp_status_init(void)
{
	int ret;

	_D("Component status init");

	ret = _request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (ret < 0) {
		_E("Failed to register commands");
		return -1;
	}

	ret = _cynara_register_checkers(__cynara_checkers,
			ARRAY_SIZE(__cynara_checkers));
	if (ret < 0) {
		_E("Failed to register cynara checkers");
		return -1;
	}

	_noti_listen(NOTI_MSG_MAIN_APP_DEAD, __on_main_app_dead);

	return 0;
}

void _comp_status_fini(void)
{
	_D("Component status fini");
	if (__comp_list)
		g_list_free_full(__comp_list, __destroy_comp_status);
}
