/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <glib.h>
#include <dirent.h>
#include <pkgmgr-info.h>
#include <aul_comp_info_internal.h>
#include <aul.h>
#include <aul_sock.h>
#include <bundle_internal.h>

#include "amd_util.h"
#include "amd_appinfo.h"
#include "amd_compinfo.h"
#include "amd_noti.h"
#include "amd_cynara.h"
#include "amd_request.h"
#include "amd_socket.h"

struct compinfo_s {
	char *val[CIT_MAX];
};

struct compinfo_localized_info_s {
	char *locale;
	char *icon;
	char *label;
};

typedef int (*compinfo_handler_add_cb)(aul_compinfo_h handle,
		struct compinfo_s *info, void *data);
typedef void (*compinfo_handler_remove_cb)(void *data);

typedef struct _compinfo_vft {
	compinfo_handler_add_cb constructor;
	compinfo_handler_remove_cb destructor;
} compinfo_vft;

struct user_compinfo_s {
	uid_t uid;
	GHashTable *tbl;
};

struct cb_info_s {
	compinfo_foreach_cb callback;
	void *user_data;
};

static GHashTable *__user_table;

static int __compinfo_add_appid(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	const char *appid;
	int ret;

	ret = aul_compinfo_get_app_id(handle, &appid);
	if (ret != 0) {
		_E("Failed to get appid");
		return -1;
	}

	info->val[CIT_APPID] = strdup(appid);
	if (!info->val[CIT_APPID]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static int __compinfo_add_taskmanage(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	int ret;
	bool is_taskmanage;

	ret = aul_compinfo_is_taskmanage(handle, &is_taskmanage);
	if (ret != 0) {
		_E("Failed to get taskmanage");
		return -1;
	}

	info->val[CIT_TASKMANAGE] =
		is_taskmanage ? strdup("true") : strdup("false");
	if (!info->val[CIT_TASKMANAGE]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static int __compinfo_add_type(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	const char *type;
	int ret;

	ret = aul_compinfo_get_type(handle, &type);
	if (ret != 0) {
		_E("Failed to get component type");
		return -1;
	}

	info->val[CIT_TYPE] = strdup(type);
	if (!info->val[CIT_TYPE]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static int __compinfo_add_launch_mode(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	const char *launch_mode;
	int ret;

	ret = aul_compinfo_get_launch_mode(handle, &launch_mode);
	if (ret != 0) {
		_E("Failed to get launch mode");
		return -1;
	}

	info->val[CIT_LAUNCH_MODE] = strdup(launch_mode);
	if (!info->val[CIT_LAUNCH_MODE]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static int __compinfo_add_main(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	bool is_main = false;
	int ret;

	ret = aul_compinfo_is_main_comp(handle, &is_main);
	if (ret != 0) {
		_E("Failed to check whether the component is main or not");
		return -1;
	}

	info->val[CIT_MAIN] = strdup(is_main ? "true" : "false");
	if (!info->val[CIT_MAIN]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static int __compinfo_add_icon_display(aul_compinfo_h handle,
		struct compinfo_s *info, void *data)
{
	bool icon_display = false;
	int ret;

	ret = aul_compinfo_is_icon_display(handle, &icon_display);
	if (ret != 0) {
		_E("Failed to check the icon display");
		return -1;
	}

	info->val[CIT_ICON_DISPLAY] = strdup(icon_display ? "true" : "false");
	if (!info->val[CIT_ICON_DISPLAY]) {
		_E("Out of memory");
		return -1;
	}

	return 0;
}

static compinfo_vft compinfo_table[CIT_MAX] = {
	[CIT_NAME] = {
		.constructor = NULL,
		.destructor = NULL
	},
	[CIT_APPID] = {
		.constructor = __compinfo_add_appid,
		.destructor = free
	},
	[CIT_TASKMANAGE] = {
		.constructor = __compinfo_add_taskmanage,
		.destructor = free
	},
	[CIT_TYPE] = {
		.constructor = __compinfo_add_type,
		.destructor = free
	},
	[CIT_LAUNCH_MODE] = {
		.constructor = __compinfo_add_launch_mode,
		.destructor = free
	},
	[CIT_MAIN] = {
		.constructor = __compinfo_add_main,
		.destructor = free
	},
	[CIT_ICON_DISPLAY] = {
		.constructor = __compinfo_add_icon_display,
		.destructor = free
	},
};

static void __compinfo_remove_handler(gpointer data)
{
	struct compinfo_s *info = (struct compinfo_s *)data;
	int i;

	if (!info)
		return;

	for (i = CIT_START; i < CIT_MAX; i++) {
		if (compinfo_table[i].destructor && info->val[i] != NULL)
			compinfo_table[i].destructor(info->val[i]);
	}

	free(info);
}

static bool __compinfo_insert_handler(aul_compinfo_h handle, void *data)
{
	struct user_compinfo_s *user_info = (struct user_compinfo_s *)data;
	struct compinfo_s *info;
	const char *comp_id;
	int r;
	int i;

	if (!handle || !user_info) {
		_E("Invalid parameter");
		return false;
	}

	r = aul_compinfo_get_comp_id(handle, &comp_id);
	if (r != 0) {
		_E("Failed to get component ID");
		return false;
	}

	g_hash_table_remove(user_info->tbl, comp_id);

	info = calloc(1, sizeof(struct compinfo_s));
	if (!info) {
		_E("Out of memory");
		return false;
	}

	info->val[CIT_NAME] = strdup(comp_id);
	if (!info->val[CIT_NAME]) {
		_E("Failed to duplicate component ID(%s)", comp_id);
		free(info);
		return false;
	}

	for (i = CIT_START; i < CIT_MAX; i++) {
		if (compinfo_table[i].constructor) {
			r = compinfo_table[i].constructor(handle, info,
					user_info);
			if (r < 0) {
				_E("Failed to load component info %s", comp_id);
				__compinfo_remove_handler(info);
				return true;
			}
		}
	}

	SECURE_LOGD("%s : %s : %s : %s",
			info->val[CIT_APPID],
			info->val[CIT_NAME],
			info->val[CIT_TYPE],
			info->val[CIT_TASKMANAGE]);

	g_hash_table_insert(user_info->tbl, info->val[CIT_NAME], info);

	return true;
}

compinfo_h _compinfo_find(uid_t uid, const char *id)
{
	struct user_compinfo_s *user_info;

	if (!id) {
		_E("Invalid parameter");
		return NULL;
	}

	user_info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (!user_info) {
		_E("Failed to find user(%u) info", uid);
		return NULL;
	}

	return (struct compinfo_s *)g_hash_table_lookup(user_info->tbl, id);
}

const char *_compinfo_get_value(compinfo_h handle, compinfo_type type)
{
	struct compinfo_s *info = (struct compinfo_s *)handle;

	if (!info) {
		_E("Invalid parameter");
		return NULL;
	}

	if (type >= CIT_MAX) {
		_E("Invalid parameter");
		return NULL;
	}

	return info->val[type];
}

static void __foreach_cb(gpointer key, gpointer value, gpointer user_data)
{
	struct compinfo_s *info = (struct compinfo_s *)value;
	struct cb_info_s *cb_info = (struct cb_info_s *)user_data;

	cb_info->callback(info, info->val[CIT_APPID],
			info->val[CIT_NAME], cb_info->user_data);
}

int _compinfo_foreach(uid_t uid, compinfo_foreach_cb callback,
		void *user_data)
{
	struct user_compinfo_s *user_info;
	struct cb_info_s cb_info = {callback, user_data};

	if (!callback) {
		_E("Invalid parameter");
		return -1;
	}

	user_info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (!user_info) {
		_E("Failed to find user(%u) info", uid);
		return -1;
	}

	g_hash_table_foreach(user_info->tbl, __foreach_cb, &cb_info);

	return 0;
}

static int __on_appinfo_insert(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	pkgmgrinfo_appinfo_h handle = (pkgmgrinfo_appinfo_h)arg3;
	struct user_compinfo_s *user_info;
	char *comp_type = NULL;
	char *appid = NULL;
	uid_t uid = (uid_t)arg1;
	int r;

	r = pkgmgrinfo_appinfo_get_component_type(handle, &comp_type);
	if (r != PMINFO_R_OK) {
		_E("Failed to get component type");
		return NOTI_CONTINUE;
	}

	if (comp_type && strcmp(comp_type, APP_TYPE_COMPONENT_BASED) != 0)
		return NOTI_CONTINUE;

	user_info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (!user_info) {
		_E("Failed to find user(%u) table", uid);
		return NOTI_CONTINUE;
	}

	r = pkgmgrinfo_appinfo_get_appid(handle, &appid);
	if (r != PMINFO_R_OK) {
		_E("Failed to get appid");
		return NOTI_CONTINUE;
	}

	r = aul_compinfo_usr_foreach_compinfo_from_app(appid, uid,
			__compinfo_insert_handler, user_info);
	if (r != 0) {
		_E("Failed to retrieve component info");
		return NOTI_CONTINUE;
	}

	return NOTI_CONTINUE;
}

static gboolean __foreach_remove_cb(gpointer key, gpointer value,
		gpointer user_data)
{
	const char *appid = (const char *)user_data;
	struct compinfo_s *info = (struct compinfo_s *)value;

	if (!strcmp(info->val[CIT_APPID], appid))
		return TRUE;

	return FALSE;
}

static int __on_appinfo_remove(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	struct user_compinfo_s *user_info;
	uid_t uid = (uid_t)arg1;
	char *appid = (char *)arg3;

	user_info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (!user_info) {
		_E("Failed to find user(%u) info", uid);
		return NOTI_CONTINUE;
	}

	g_hash_table_foreach_remove(user_info->tbl, __foreach_remove_cb, appid);

	return NOTI_CONTINUE;
}

static struct user_compinfo_s *__create_user_compinfo(uid_t uid)
{
	struct user_compinfo_s *info;

	info = calloc(1, sizeof(struct user_compinfo_s));
	if (!info) {
		_E("Out of memory");
		return NULL;
	}

	info->uid = uid;
	info->tbl = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
			__compinfo_remove_handler);
	if (!info->tbl) {
		_E("Out of memory");
		free(info);
		return NULL;
	}

	return info;
}

static void __destroy_user_compinfo(gpointer data)
{
	struct user_compinfo_s *info = (struct user_compinfo_s *)data;

	if (!info)
		return;

	if (info->tbl)
		g_hash_table_destroy(info->tbl);
	free(info);
}

static int __on_login_monitor_login_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	struct user_compinfo_s *info;
	uid_t uid = (uid_t)arg1;

	info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (info)
		return NOTI_CONTINUE;

	info = __create_user_compinfo(uid);
	if (!info) {
		_E("Failed to create user(%u) info", uid);
		return NOTI_CONTINUE;
	}

	g_hash_table_insert(__user_table, GUINT_TO_POINTER(uid), info);

	return NOTI_CONTINUE;
}

static int __on_login_monitor_logout(const char *msg, int arg1, int arg2,
		void *arg3, bundle *b)
{
	struct user_compinfo_s *info;
	uid_t uid = (uid_t)arg1;

	info = (struct user_compinfo_s *)g_hash_table_lookup(__user_table,
			GUINT_TO_POINTER(uid));
	if (!info)
		return NOTI_CONTINUE;

	g_hash_table_remove(__user_table, GUINT_TO_POINTER(uid));

	return NOTI_CONTINUE;
}

int _compinfo_init(void)
{
	_D("Component info init");

	__user_table = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, __destroy_user_compinfo);
	if (!__user_table) {
		_E("Failed to create user table");
		return -1;
	}

	_noti_listen(NOTI_MSG_LOGIN_MONITOR_LOGIN_START,
			__on_login_monitor_login_start);
	_noti_listen(NOTI_MSG_LOGIN_MONITOR_LOGOUT,
			__on_login_monitor_logout);
	_noti_listen(NOTI_MSG_APPINFO_INSERT,
			__on_appinfo_insert);
	_noti_listen(NOTI_MSG_APPINFO_REMOVE,
			__on_appinfo_remove);

	return 0;
}

void _compinfo_fini(void)
{
	_D("Component info fini");

	if (__user_table)
		g_hash_table_destroy(__user_table);
}
