/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <bundle_internal.h>
#include <aul.h>
#include <aul_cmd.h>
#include <aul_sock.h>

#include "amd_util.h"
#include "amd_app_com.h"
#include "amd_noti.h"
#include "amd_request.h"
#include "amd_app_status.h"
#include "amd_launcher_service.h"

static bundle *__create_bundle(const char *app_id, const char *instance_id,
		int pid, const char *serial)
{
	char pid_str[MAX_PID_STR_BUFSZ];
	bundle *b;

	b = bundle_create();
	if (!b) {
		_E("Out of memory");
		return NULL;
	}

	bundle_add(b, AUL_K_APPID, app_id);
	bundle_add(b, AUL_K_INSTANCE_ID, instance_id ? instance_id : "");

	snprintf(pid_str, sizeof(pid_str), "%d", pid);
	bundle_add(b, AUL_K_PID, pid_str);
	bundle_add(b, AUL_K_LAUNCHER_SERVICE_SERIAL, serial);

	return b;
}

static void __publish_launcher_service(const char *endpoint,
		const char *app_id, const char *instance_id, int pid,
		const char *serial, uid_t uid)
{
	bundle *b;

	b = __create_bundle(app_id, instance_id, pid, serial);
	if (!b)
		return;

	_app_com_send(endpoint, pid, b, uid);
	bundle_free(b);
}

static const char *__get_instance_id(int pid)
{
	app_status_h app_status;

	app_status = _app_status_find_v2(pid);
	if (!app_status)
		return NULL;

	return _app_status_get_instance_id(app_status);
}

static int __on_launch_do_starting_app_end(const char *msg,
		int arg1, int arg2, void *arg3, bundle *data)
{
	int pid = arg1;
	int bg_launch = arg2;
	request_h req = (request_h)data;
	bundle *b = _request_get_bundle(req);
	int caller_pid = _request_get_pid(req);
	uid_t uid = _request_get_uid(req);
	char endpoint[MAX_LOCAL_BUFSZ];
	const char *instance_id;
	const char *app_id;
	const char *launcher_service;
	const char *serial;

	if (pid < 0 || uid < REGULAR_UID_MIN)
		return NOTI_CONTINUE;

	if (bg_launch) {
		_W("Background launch");
		return NOTI_CONTINUE;
	}

	launcher_service = bundle_get_val(b, AUL_K_LAUNCHER_SERVICE);
	if (!launcher_service)
		return NOTI_CONTINUE;

	serial = bundle_get_val(b, AUL_K_LAUNCHER_SERVICE_SERIAL);
	if (!serial) {
		_E("Failed to get serial");
		return NOTI_CONTINUE;
	}

	snprintf(endpoint, sizeof(endpoint), "launcher_service:%s:%d",
			launcher_service, caller_pid);
	if (!_app_com_endpoint_exists(endpoint))
		return NOTI_CONTINUE;

	app_id = bundle_get_val(b, AUL_K_APPID);
	instance_id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!instance_id)
		instance_id = __get_instance_id(pid);

	__publish_launcher_service(endpoint, app_id, instance_id, pid,
			serial, uid);

	return NOTI_CONTINUE;
}

int _launcher_service_init(void)
{
	_D("init");
	_noti_listen(NOTI_MSG_LAUNCH_DO_STARTING_APP_RELAUNCH_END,
			__on_launch_do_starting_app_end);
	_noti_listen(NOTI_MSG_LAUNCH_DO_STARTING_APP_END,
			__on_launch_do_starting_app_end);

	return 0;
}

void _launcher_service_fini(void)
{
	_D("fini");
}
