/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <gio/gio.h>
#include <glib.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <unistd.h>

#include "amd_receiver.h"
#include "amd_util.h"

typedef struct receiver_s {
	GThread *thread;
	GMutex mutex;
	GCond cond;
	GMainContext *context;
	GMainLoop *loop;
	GIOChannel *channel;
	guint tag;
	int fd;
	receiver_cb callback;
	void *user_data;
} receiver_t;

static receiver_t *__get_inst(void)
{
	static receiver_t inst;

	return &inst;
}

static void __worker_unset_io_handler(void)
{
	receiver_t *inst = __get_inst();
	GSource *source;

	if (inst->tag) {
		source = g_main_context_find_source_by_id(NULL, inst->tag);
		if (source && !g_source_is_destroyed(source))
			g_source_remove(inst->tag);
	}

	if (inst->channel)
		g_io_channel_unref(inst->channel);
}

static int __set_comm(const char *name)
{
	int fd;
	ssize_t bytes_written;
	char path[PATH_MAX];
	pid_t tid = syscall(__NR_gettid);

	_I("[%s] TID(%d)", name, tid);
	snprintf(path, sizeof(path), "/proc/%d/comm", tid);
	fd = open(path, O_WRONLY);
	if (fd < 0) {
		_E("Failed to open %s. error(%d)", path, errno);
		return -1;
	}

	bytes_written = write(fd, name, strlen(name) + 1);
	if (bytes_written < 0) {
		_E("Failed to write name(%s)", name);
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

static gboolean __io_handler(GIOChannel *io, GIOCondition condition,
		gpointer data)
{
	int fd = g_io_channel_unix_get_fd(io);
	receiver_t *inst = __get_inst();
	receiver_cb callback = inst->callback;
	void *user_data = inst->user_data;

	if (callback && callback(fd, user_data))
		return G_SOURCE_CONTINUE;

	inst->tag = 0;
	return G_SOURCE_REMOVE;
}

static int __worker_set_io_handler(int fd)
{
	GIOCondition cond = G_IO_IN | G_IO_PRI | G_IO_HUP | G_IO_ERR;
	receiver_t *inst = __get_inst();
	GIOChannel *channel;
	GSource *source;

	if (fd < 3) {
		_E("Invalid parameter");
		return -EINVAL;
	}

	channel = g_io_channel_unix_new(fd);
	if (!channel) {
		_E("Failed to create GIOChannel");
		return -ENOMEM;
	}

	source = g_io_create_watch(channel, cond);
	if (!source) {
		_E("Failed to create GSource");
		g_io_channel_unref(channel);
		return -ENOMEM;
	}

	g_source_set_callback(source, (GSourceFunc)__io_handler, inst, NULL);
	g_source_set_priority(source, G_PRIORITY_DEFAULT);

	g_mutex_lock(&inst->mutex);
	inst->channel = channel;
	inst->tag = g_source_attach(source, inst->context);
	g_mutex_unlock(&inst->mutex);

	g_source_unref(source);

	return 0;
}

static void __worker_destroy(void)
{
	receiver_t *inst = __get_inst();

	if (inst->thread) {
		g_mutex_lock(&inst->mutex);
		if (g_main_loop_is_running(inst->loop)) {
			g_main_loop_quit(inst->loop);
			g_cond_wait(&inst->cond, &inst->mutex);
		} else {
			_E("GMainLoop is not running");
		}
		g_mutex_unlock(&inst->mutex);
		g_thread_join(inst->thread);
	}

	g_cond_clear(&inst->cond);
	g_mutex_clear(&inst->mutex);
}

static gboolean __notify_cb(gpointer data)
{
	receiver_t *inst = __get_inst();

	_W("GMainLoop is started");
	g_mutex_lock(&inst->mutex);
	g_cond_signal(&inst->cond);
	g_mutex_unlock(&inst->mutex);

	return G_SOURCE_REMOVE;
}

static gpointer __worker_thread_loop(gpointer data)
{
	receiver_t *inst = __get_inst();
	GSource *source;

	g_mutex_lock(&inst->mutex);
	__set_comm("receiver+");

	source = g_idle_source_new();
	if (!source) {
		_E("Failed to create GSource");
		g_cond_signal(&inst->cond);
		g_mutex_unlock(&inst->mutex);
		return NULL;
	}

	inst->context = g_main_context_new();
	if (!inst->context) {
		_E("Failed to create GMainContext");
		g_source_unref(source);
		g_cond_signal(&inst->cond);
		g_mutex_unlock(&inst->mutex);
		return NULL;
	}

	g_source_set_callback(source, (GSourceFunc)__notify_cb, inst, NULL);
	g_source_set_priority(source, G_PRIORITY_HIGH);
	g_source_attach(source, inst->context);
	g_source_unref(source);

	inst->loop = g_main_loop_new(inst->context, FALSE);
	g_main_context_push_thread_default(inst->context);
	g_mutex_unlock(&inst->mutex);

	if (!inst->loop) {
		_E("Failed to create GMainLoop");
		goto end;
	}

	g_main_loop_run(inst->loop);
end:
	g_mutex_lock(&inst->mutex);
	g_main_context_pop_thread_default(inst->context);
	g_main_loop_unref(inst->loop);
	g_main_context_unref(inst->context);
	g_cond_signal(&inst->cond);
	g_mutex_unlock(&inst->mutex);
	_W("Shut down worker");

	return NULL;
}

static int __worker_create(void)
{
	receiver_t *inst = __get_inst();

	g_mutex_init(&inst->mutex);
	g_cond_init(&inst->cond);

	g_mutex_lock(&inst->mutex);

	inst->thread = g_thread_new("receiver+", __worker_thread_loop, inst);
	if (!inst->thread) {
		_E("Failed to create worker thread");
		return -1;
	}

	g_cond_wait(&inst->cond, &inst->mutex);
	g_mutex_unlock(&inst->mutex);

	return 0;
}

int _receiver_set_handler(int fd, receiver_cb callback, void *user_data)
{
	receiver_t *inst = __get_inst();
	int ret;

	g_mutex_lock(&inst->mutex);
	inst->fd = fd;
	inst->callback = callback;
	inst->user_data = user_data;
	g_mutex_unlock(&inst->mutex);

	ret = __worker_set_io_handler(fd);
	if (ret < 0) {
		_E("Failed to set io handler");
		return ret;
	}

	return 0;
}

int _receiver_init(void)
{
	int ret;

	_D("[__RECEIVER__] init");
	ret = __worker_create();
	if (ret < 0)
		return ret;

	return 0;
}

void _receiver_fini(void)
{
	_D("[__RECEIVER__] fini");
	__worker_unset_io_handler();
	__worker_destroy();
}
